"Logging utilities."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
from typing import TYPE_CHECKING, Any, MutableMapping, TypeAlias


def get_setting(name: str, default: str = "") -> bool:
    "Retrieve environment variable setting."
    value = os.environ.get(name, default)
    if not value:
        return False
    return value.strip().lower() not in {"0", "false"}


FINSY_DEBUG = get_setting("FINSY_DEBUG")


if TYPE_CHECKING:  # pragma: no cover
    _BaseLoggerAdapter: TypeAlias = logging.LoggerAdapter[logging.Logger]
else:
    # logging.LoggerAdapter will be generic at runtime in Python 3.11.
    _BaseLoggerAdapter: TypeAlias = logging.LoggerAdapter


def _get_current_task_name(shorten: bool = False) -> str:
    "Return the name of the current task (or '' if there is none.)"
    # current_task() raises RuntimeError if there is no running event loop.
    try:
        task = asyncio.current_task()
    except RuntimeError:
        return ""

    # Task may be None if we're not in a task; e.g. in a low-level callback.
    if not task:
        return ""

    # Remove the "fy:" prefix from the task name.
    name = task.get_name()
    if name.startswith("fy:"):
        name = name[3:]

    # Shorten task name to omit details after "|".
    if shorten:
        pos = name.find("|")
        if pos > 0:
            name = name[:pos]

    return name


class LoggerAdapter(_BaseLoggerAdapter):
    """Custom log adapter to include the name of the current task."""

    def process(
        self,
        msg: Any,
        kwargs: MutableMapping[str, Any],
    ) -> tuple[Any, MutableMapping[str, Any]]:
        """Process the logging message and keyword arguments passed in to a
        logging call to insert contextual information.
        """
        task_name = _get_current_task_name()
        return f"[{task_name}] {msg}", kwargs

    def info(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """INFO level uses a concise task name representation for readability."""
        if self.logger.isEnabledFor(logging.INFO):
            task_name = _get_current_task_name(True)
            self.logger.info(f"[{task_name}] {msg}", *args, **kwargs)


LOGGER = LoggerAdapter(logging.getLogger(__package__))
MSG_LOG = LoggerAdapter(logging.getLogger(f"{__package__}.msg"))

if FINSY_DEBUG:  # pragma: no cover
    # If FINSY_DEBUG is true, log all protobuf messages.
    MSG_LOG.setLevel(logging.DEBUG)



"gNMI server for testing with hard-wired responses."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import contextlib
from typing import AsyncIterator

import grpc  # pyright: ignore[reportMissingTypeStubs]

from finsy import pbutil
from finsy.log import LOGGER
from finsy.proto import gnmi, gnmi_grpc

# pyright: reportIncompatibleMethodOverride=false
# pylint: disable=invalid-overridden-method,line-too-long


class GNMIServer(gnmi_grpc.gNMIServicer):
    "Test gNMI server."

    _listen_addr: str
    _server: grpc.aio.Server | None

    def __init__(self, listen_addr: str):
        "Initialize gNMI server."
        self._listen_addr = listen_addr
        self._server = None

    @contextlib.asynccontextmanager
    async def run(self):
        "Run server inside an async context manager."
        self._server = self._create_server()
        try:
            await self._server.start()
            LOGGER.debug("GNMIServer: server started: %s", self._listen_addr)
            yield self

        finally:
            await self._server.stop(None)
            await asyncio.sleep(0.010)  # give a little time to clean up tasks
            self._server = None
            LOGGER.debug("GNMIServer: server stopped: %s", self._listen_addr)

    def _create_server(self) -> grpc.aio.Server:
        "Create AIO server."
        server = grpc.aio.server()
        gnmi_grpc.add_gNMIServicer_to_server(self, server)
        server.add_insecure_port(self._listen_addr)
        return server

    async def Get(
        self,
        request: gnmi.GetRequest,
        context: grpc.aio.ServicerContext[gnmi.GetRequest, gnmi.GetResponse],
    ) -> gnmi.GetResponse:
        "Handle gNMI GetRequest."
        if len(request.path) == 1:
            if not request.path[0].elem and request.type == gnmi.GetRequest.CONFIG:
                return pbutil.from_text(_GNMI_CONFIG_STRATUM, gnmi.GetResponse)

        return pbutil.from_text(_GNMI_INTERFACE_STUFF, gnmi.GetResponse)

    async def Set(
        self,
        request: gnmi.SetRequest,
        context: grpc.aio.ServicerContext[gnmi.SetRequest, gnmi.SetResponse],
    ) -> gnmi.SetResponse:
        "Handle the set request."
        # TODO: Does not actually do anything.
        results = (
            [
                gnmi.UpdateResult(
                    path=update.path, op=gnmi.UpdateResult.Operation.UPDATE
                )
                for update in request.update
            ]
            + [
                gnmi.UpdateResult(
                    path=update.path, op=gnmi.UpdateResult.Operation.REPLACE
                )
                for update in request.replace
            ]
            + [
                gnmi.UpdateResult(path=path, op=gnmi.UpdateResult.Operation.DELETE)
                for path in request.delete
            ]
        )

        return gnmi.SetResponse(response=results)

    async def Capabilities(
        self,
        request: gnmi.CapabilityRequest,
        context: grpc.aio.ServicerContext[
            gnmi.CapabilityRequest, gnmi.CapabilityResponse
        ],
    ) -> gnmi.CapabilityResponse:
        "Handle the capability request."
        return pbutil.from_text(_GNMI_CAPABILITIES, gnmi.CapabilityResponse)

    async def Subscribe(
        self,
        request_iterator: AsyncIterator[gnmi.SubscribeRequest],
        _context: grpc.aio.ServicerContext[
            gnmi.SubscribeRequest, gnmi.SubscribeResponse
        ],
    ) -> AsyncIterator[gnmi.SubscribeResponse]:
        "Handle the subscribe request."
        async for msg in request_iterator:
            match msg.WhichOneof("request"):
                case "subscribe":
                    mode = msg.subscribe.subscription[0].mode
                    if mode == gnmi.SubscriptionMode.SAMPLE:
                        for reply in _GNMI_SUBSCRIBE_RESPONSES_SAMPLE:
                            await asyncio.sleep(0.1)
                            yield pbutil.from_text(reply, gnmi.SubscribeResponse)
                    else:
                        for reply in _GNMI_SUBSCRIBE_RESPONSES_INITIAL:
                            yield pbutil.from_text(reply, gnmi.SubscribeResponse)
                        yield gnmi.SubscribeResponse(sync_response=True)
                        if msg.subscribe.mode != gnmi.SubscriptionList.ONCE:
                            await asyncio.sleep(0.2)
                            for reply in _GNMI_SUBSCRIBE_RESPONSES_UPDATES:
                                yield pbutil.from_text(reply, gnmi.SubscribeResponse)
                case "poll":  # not supported
                    raise RuntimeError("not supported")
                case _:
                    pass


# Copy and pasted from Stratum responses (minor editing).

_GNMI_CONFIG_STRATUM = r"""
notification {
  update {
    path {
    }
    val {
      any_val {
        type_url: "type.googleapis.com/openconfig.Device"
        value: "\252\221\231\304\001\025\n\002s1\022\017\242\313\237\312\004\t\250\314\333\264\010\261\366\216\006\252\221\231\304\001 \n\005:lc-1\022\027\202\377\211H\t\322\315\356\351\007\003\n\0011\212\212\344\255\007\003\n\0011\252\221\231\304\001R\n\007s1-eth1\022G\202\377\211H\t\322\315\356\351\007\003\n\0011\322\374\322\233\001\010\372\375\371\331\n\002\010\001\312\212\243\352\005\010\312\223\360\266\005\002\010\001\362\262\360\266\010\t\n\007s1-eth1\312\261\312\343\017\010\262\250\370\240\002\002\010\001\252\221\231\304\001R\n\007s1-eth2\022G\202\377\211H\t\322\315\356\351\007\003\n\0011\322\374\322\233\001\010\372\375\371\331\n\002\010\001\312\212\243\352\005\010\312\223\360\266\005\002\010\001\362\262\360\266\010\t\n\007s1-eth2\312\261\312\343\017\010\262\250\370\240\002\002\010\002\322\277\322\313\0148\n\007s1-eth1\022-\352\366\377\215\001\002\010\001\262\241\245\260\001\002\010\001\242\207\355\257\002\017\330\220\301\217\010\304\371\266i\322\354\274\350\n\000\352\340\372\222\003\002\010\001\322\277\322\313\0148\n\007s1-eth2\022-\352\366\377\215\001\002\010\001\262\241\245\260\001\002\010\001\242\207\355\257\002\017\330\220\301\217\010\304\371\266i\322\354\274\350\n\000\352\340\372\222\003\002\010\002"
      }
    }
  }
}
"""

_GNMI_CAPABILITIES = r"""
supported_models { name: "openconfig-interfaces-stratum" organization: "Open Networking Foundation" version: "0.1.0" }
supported_encodings: PROTO
gNMI_version: "0.7.0"
"""


_GNMI_INTERFACE_STUFF = r"""
notification {
  timestamp: 1656719438871244764
  update {
    path {
      elem {
        name: "interfaces"
      }
      elem {
        name: "interface"
        key {
          key: "name"
          value: "s1-eth1"
        }
      }
      elem {
        name: "state"
      }
      elem {
        name: "ifindex"
      }
    }
    val {
      uint_val: 1
    }
  }
}
notification {
  timestamp: 1656719438871263864
  update {
    path {
      elem {
        name: "interfaces"
      }
      elem {
        name: "interface"
        key {
          key: "name"
          value: "s1-eth2"
        }
      }
      elem {
        name: "state"
      }
      elem {
        name: "ifindex"
      }
    }
    val {
      uint_val: 2
    }
  }
}
"""


_GNMI_SUBSCRIBE_RESPONSES_INITIAL = [
    r"""
    update { timestamp: 1656723692816736079 update { path { elem { name: "interfaces" } elem { name: "interface" key { key: "name" value: "s1-eth1" } } elem { name: "state" } elem { name: "oper-status" } } val { string_val: "UP" } } }
    """,
    r"""
    update { timestamp: 1656723692817152380 update { path { elem { name: "interfaces" } elem { name: "interface" key { key: "name" value: "s1-eth2" } } elem { name: "state" } elem { name: "oper-status" } } val { string_val: "UP" } } }
    """,
]


_GNMI_SUBSCRIBE_RESPONSES_UPDATES = [
    r"""
    update { timestamp: 1656723692816736079 update { path { elem { name: "interfaces" } elem { name: "interface" key { key: "name" value: "s1-eth1" } } elem { name: "state" } elem { name: "oper-status" } } val { string_val: "DOWN" } } }
    """,
    r"""
    update { timestamp: 1656723692816736079 update { path { elem { name: "interfaces" } elem { name: "interface" key { key: "name" value: "s1-eth1" } } elem { name: "state" } elem { name: "oper-status" } } val { string_val: "UP" } } }
    """,
]


_GNMI_SUBSCRIBE_RESPONSES_SAMPLE = [
    r"""
    update { timestamp: 1656728002678626415 update { path { elem { name: "interfaces" } elem { name: "interface" key { key: "name" value: "s1-eth1" } } elem { name: "state" } elem { name: "counters" } elem { name: "in-octets" } } val { uint_val: 1196 } } }
    """,
    r"""
    update { timestamp: 1656728002680768004 update { path { elem { name: "interfaces" } elem { name: "interface" key { key: "name" value: "s1-eth2" } } elem { name: "state" } elem { name: "counters" } elem { name: "in-octets" } } val { uint_val: 1196 } } }
    """,
    r"""
    update { timestamp: 1656728002779297318 update { path { elem { name: "interfaces" } elem { name: "interface" key { key: "name" value: "s1-eth1" } } elem { name: "state" } elem { name: "counters" } elem { name: "in-octets" } } val { uint_val: 1196 } } }
    """,
    r"""
    update { timestamp: 1656728002782083970 update { path { elem { name: "interfaces" } elem { name: "interface" key { key: "name" value: "s1-eth2" } } elem { name: "state" } elem { name: "counters" } elem { name: "in-octets" } } val { uint_val: 1196 } } }
    """,
    r"""
    update { timestamp: 1656728002879090194 update { path { elem { name: "interfaces" } elem { name: "interface" key { key: "name" value: "s1-eth1" } } elem { name: "state" } elem { name: "counters" } elem { name: "in-octets" } } val { uint_val: 1196 } } }
    """,
]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.struct_pb2
import google.protobuf.timestamp_pb2
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class AttributeContext(google.protobuf.message.Message):
    """This message defines the standard attribute vocabulary for Google APIs.

    An attribute is a piece of metadata that describes an activity on a network
    service. For example, the size of an HTTP request, or the status code of
    an HTTP response.

    Each attribute has a type and a name, which is logically defined as
    a proto message field in `AttributeContext`. The field type becomes the
    attribute type, and the field path becomes the attribute name. For example,
    the attribute `source.ip` maps to field `AttributeContext.source.ip`.

    This message definition is guaranteed not to have any wire breaking change.
    So you can use it directly for passing attributes across different systems.

    NOTE: Different system may generate different subset of attributes. Please
    verify the system specification before relying on an attribute generated
    a system.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Peer(google.protobuf.message.Message):
        """This message defines attributes for a node that handles a network request.
        The node can be either a service or an application that sends, forwards,
        or receives the request. Service peers should fill in
        `principal` and `labels` as appropriate.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class LabelsEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            value: builtins.str
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: builtins.str = ...,
            ) -> None: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        IP_FIELD_NUMBER: builtins.int
        PORT_FIELD_NUMBER: builtins.int
        LABELS_FIELD_NUMBER: builtins.int
        PRINCIPAL_FIELD_NUMBER: builtins.int
        REGION_CODE_FIELD_NUMBER: builtins.int
        ip: builtins.str
        """The IP address of the peer."""
        port: builtins.int
        """The network port of the peer."""
        principal: builtins.str
        """The identity of this peer. Similar to `Request.auth.principal`, but
        relative to the peer instead of the request. For example, the
        idenity associated with a load balancer that forwared the request.
        """
        region_code: builtins.str
        """The CLDR country/region code associated with the above IP address.
        If the IP address is private, the `region_code` should reflect the
        physical location where this peer is running.
        """
        @property
        def labels(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
            """The labels associated with the peer."""

        def __init__(
            self,
            *,
            ip: builtins.str = ...,
            port: builtins.int = ...,
            labels: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
            principal: builtins.str = ...,
            region_code: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["ip", b"ip", "labels", b"labels", "port", b"port", "principal", b"principal", "region_code", b"region_code"]) -> None: ...

    @typing.final
    class Api(google.protobuf.message.Message):
        """This message defines attributes associated with API operations, such as
        a network API request. The terminology is based on the conventions used
        by Google APIs, Istio, and OpenAPI.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SERVICE_FIELD_NUMBER: builtins.int
        OPERATION_FIELD_NUMBER: builtins.int
        PROTOCOL_FIELD_NUMBER: builtins.int
        VERSION_FIELD_NUMBER: builtins.int
        service: builtins.str
        """The API service name. It is a logical identifier for a networked API,
        such as "pubsub.googleapis.com". The naming syntax depends on the
        API management system being used for handling the request.
        """
        operation: builtins.str
        """The API operation name. For gRPC requests, it is the fully qualified API
        method name, such as "google.pubsub.v1.Publisher.Publish". For OpenAPI
        requests, it is the `operationId`, such as "getPet".
        """
        protocol: builtins.str
        """The API protocol used for sending the request, such as "http", "https",
        "grpc", or "internal".
        """
        version: builtins.str
        """The API version associated with the API operation above, such as "v1" or
        "v1alpha1".
        """
        def __init__(
            self,
            *,
            service: builtins.str = ...,
            operation: builtins.str = ...,
            protocol: builtins.str = ...,
            version: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["operation", b"operation", "protocol", b"protocol", "service", b"service", "version", b"version"]) -> None: ...

    @typing.final
    class Auth(google.protobuf.message.Message):
        """This message defines request authentication attributes. Terminology is
        based on the JSON Web Token (JWT) standard, but the terms also
        correlate to concepts in other standards.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        PRINCIPAL_FIELD_NUMBER: builtins.int
        AUDIENCES_FIELD_NUMBER: builtins.int
        PRESENTER_FIELD_NUMBER: builtins.int
        CLAIMS_FIELD_NUMBER: builtins.int
        ACCESS_LEVELS_FIELD_NUMBER: builtins.int
        principal: builtins.str
        """The authenticated principal. Reflects the issuer (`iss`) and subject
        (`sub`) claims within a JWT. The issuer and subject should be `/`
        delimited, with `/` percent-encoded within the subject fragment. For
        Google accounts, the principal format is:
        "https://accounts.google.com/{id}"
        """
        presenter: builtins.str
        """The authorized presenter of the credential. Reflects the optional
        Authorized Presenter (`azp`) claim within a JWT or the
        OAuth client id. For example, a Google Cloud Platform client id looks
        as follows: "123456789012.apps.googleusercontent.com".
        """
        @property
        def audiences(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
            """The intended audience(s) for this authentication information. Reflects
            the audience (`aud`) claim within a JWT. The audience
            value(s) depends on the `issuer`, but typically include one or more of
            the following pieces of information:

            *  The services intended to receive the credential such as
               ["pubsub.googleapis.com", "storage.googleapis.com"]
            *  A set of service-based scopes. For example,
               ["https://www.googleapis.com/auth/cloud-platform"]
            *  The client id of an app, such as the Firebase project id for JWTs
               from Firebase Auth.

            Consult the documentation for the credential issuer to determine the
            information provided.
            """

        @property
        def claims(self) -> google.protobuf.struct_pb2.Struct:
            """Structured claims presented with the credential. JWTs include
            `{key: value}` pairs for standard and private claims. The following
            is a subset of the standard required and optional claims that would
            typically be presented for a Google-based JWT:

               {'iss': 'accounts.google.com',
                'sub': '113289723416554971153',
                'aud': ['123456789012', 'pubsub.googleapis.com'],
                'azp': '123456789012.apps.googleusercontent.com',
                'email': 'jsmith@example.com',
                'iat': 1353601026,
                'exp': 1353604926}

            SAML assertions are similarly specified, but with an identity provider
            dependent structure.
            """

        @property
        def access_levels(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
            """A list of access level resource names that allow resources to be
            accessed by authenticated requester. It is part of Secure GCP processing
            for the incoming request. An access level string has the format:
            "//{api_service_name}/accessPolicies/{policy_id}/accessLevels/{short_name}"

            Example:
            "//accesscontextmanager.googleapis.com/accessPolicies/MY_POLICY_ID/accessLevels/MY_LEVEL"
            """

        def __init__(
            self,
            *,
            principal: builtins.str = ...,
            audiences: collections.abc.Iterable[builtins.str] | None = ...,
            presenter: builtins.str = ...,
            claims: google.protobuf.struct_pb2.Struct | None = ...,
            access_levels: collections.abc.Iterable[builtins.str] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["claims", b"claims"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["access_levels", b"access_levels", "audiences", b"audiences", "claims", b"claims", "presenter", b"presenter", "principal", b"principal"]) -> None: ...

    @typing.final
    class Request(google.protobuf.message.Message):
        """This message defines attributes for an HTTP request. If the actual
        request is not an HTTP request, the runtime system should try to map
        the actual request to an equivalent HTTP request.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class HeadersEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            value: builtins.str
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: builtins.str = ...,
            ) -> None: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        ID_FIELD_NUMBER: builtins.int
        METHOD_FIELD_NUMBER: builtins.int
        HEADERS_FIELD_NUMBER: builtins.int
        PATH_FIELD_NUMBER: builtins.int
        HOST_FIELD_NUMBER: builtins.int
        SCHEME_FIELD_NUMBER: builtins.int
        QUERY_FIELD_NUMBER: builtins.int
        TIME_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        PROTOCOL_FIELD_NUMBER: builtins.int
        REASON_FIELD_NUMBER: builtins.int
        AUTH_FIELD_NUMBER: builtins.int
        id: builtins.str
        """The unique ID for a request, which can be propagated to downstream
        systems. The ID should have low probability of collision
        within a single day for a specific service.
        """
        method: builtins.str
        """The HTTP request method, such as `GET`, `POST`."""
        path: builtins.str
        """The HTTP URL path."""
        host: builtins.str
        """The HTTP request `Host` header value."""
        scheme: builtins.str
        """The HTTP URL scheme, such as `http` and `https`."""
        query: builtins.str
        """The HTTP URL query in the format of `name1=value1&name2=value2`, as it
        appears in the first line of the HTTP request. No decoding is performed.
        """
        size: builtins.int
        """The HTTP request size in bytes. If unknown, it must be -1."""
        protocol: builtins.str
        """The network protocol used with the request, such as "http/1.1",
        "spdy/3", "h2", "h2c", "webrtc", "tcp", "udp", "quic". See
        https://www.iana.org/assignments/tls-extensiontype-values/tls-extensiontype-values.xhtml#alpn-protocol-ids
        for details.
        """
        reason: builtins.str
        """A special parameter for request reason. It is used by security systems
        to associate auditing information with a request.
        """
        @property
        def headers(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
            """The HTTP request headers. If multiple headers share the same key, they
            must be merged according to the HTTP spec. All header keys must be
            lowercased, because HTTP header keys are case-insensitive.
            """

        @property
        def time(self) -> google.protobuf.timestamp_pb2.Timestamp:
            """The timestamp when the `destination` service receives the first byte of
            the request.
            """

        @property
        def auth(self) -> global___AttributeContext.Auth:
            """The request authentication. May be absent for unauthenticated requests.
            Derived from the HTTP request `Authorization` header or equivalent.
            """

        def __init__(
            self,
            *,
            id: builtins.str = ...,
            method: builtins.str = ...,
            headers: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
            path: builtins.str = ...,
            host: builtins.str = ...,
            scheme: builtins.str = ...,
            query: builtins.str = ...,
            time: google.protobuf.timestamp_pb2.Timestamp | None = ...,
            size: builtins.int = ...,
            protocol: builtins.str = ...,
            reason: builtins.str = ...,
            auth: global___AttributeContext.Auth | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["auth", b"auth", "time", b"time"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["auth", b"auth", "headers", b"headers", "host", b"host", "id", b"id", "method", b"method", "path", b"path", "protocol", b"protocol", "query", b"query", "reason", b"reason", "scheme", b"scheme", "size", b"size", "time", b"time"]) -> None: ...

    @typing.final
    class Response(google.protobuf.message.Message):
        """This message defines attributes for a typical network response. It
        generally models semantics of an HTTP response.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class HeadersEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            value: builtins.str
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: builtins.str = ...,
            ) -> None: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        CODE_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        HEADERS_FIELD_NUMBER: builtins.int
        TIME_FIELD_NUMBER: builtins.int
        code: builtins.int
        """The HTTP response status code, such as `200` and `404`."""
        size: builtins.int
        """The HTTP response size in bytes. If unknown, it must be -1."""
        @property
        def headers(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
            """The HTTP response headers. If multiple headers share the same key, they
            must be merged according to HTTP spec. All header keys must be
            lowercased, because HTTP header keys are case-insensitive.
            """

        @property
        def time(self) -> google.protobuf.timestamp_pb2.Timestamp:
            """The timestamp when the `destination` service generates the first byte of
            the response.
            """

        def __init__(
            self,
            *,
            code: builtins.int = ...,
            size: builtins.int = ...,
            headers: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
            time: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["time", b"time"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["code", b"code", "headers", b"headers", "size", b"size", "time", b"time"]) -> None: ...

    @typing.final
    class Resource(google.protobuf.message.Message):
        """This message defines core attributes for a resource. A resource is an
        addressable (named) entity provided by the destination service. For
        example, a file stored on a network storage service.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class LabelsEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            value: builtins.str
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: builtins.str = ...,
            ) -> None: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        SERVICE_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        TYPE_FIELD_NUMBER: builtins.int
        LABELS_FIELD_NUMBER: builtins.int
        service: builtins.str
        """The name of the service that this resource belongs to, such as
        `pubsub.googleapis.com`. The service may be different from the DNS
        hostname that actually serves the request.
        """
        name: builtins.str
        """The stable identifier (name) of a resource on the `service`. A resource
        can be logically identified as "//{resource.service}/{resource.name}".
        The differences between a resource name and a URI are:

        *   Resource name is a logical identifier, independent of network
            protocol and API version. For example,
            `//pubsub.googleapis.com/projects/123/topics/news-feed`.
        *   URI often includes protocol and version information, so it can
            be used directly by applications. For example,
            `https://pubsub.googleapis.com/v1/projects/123/topics/news-feed`.

        See https://cloud.google.com/apis/design/resource_names for details.
        """
        type: builtins.str
        """The type of the resource. The syntax is platform-specific because
        different platforms define their resources differently.

        For Google APIs, the type format must be "{service}/{kind}".
        """
        @property
        def labels(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
            """The labels or tags on the resource, such as AWS resource tags and
            Kubernetes resource labels.
            """

        def __init__(
            self,
            *,
            service: builtins.str = ...,
            name: builtins.str = ...,
            type: builtins.str = ...,
            labels: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["labels", b"labels", "name", b"name", "service", b"service", "type", b"type"]) -> None: ...

    ORIGIN_FIELD_NUMBER: builtins.int
    SOURCE_FIELD_NUMBER: builtins.int
    DESTINATION_FIELD_NUMBER: builtins.int
    REQUEST_FIELD_NUMBER: builtins.int
    RESPONSE_FIELD_NUMBER: builtins.int
    RESOURCE_FIELD_NUMBER: builtins.int
    API_FIELD_NUMBER: builtins.int
    @property
    def origin(self) -> global___AttributeContext.Peer:
        """The origin of a network activity. In a multi hop network activity,
        the origin represents the sender of the first hop. For the first hop,
        the `source` and the `origin` must have the same content.
        """

    @property
    def source(self) -> global___AttributeContext.Peer:
        """The source of a network activity, such as starting a TCP connection.
        In a multi hop network activity, the source represents the sender of the
        last hop.
        """

    @property
    def destination(self) -> global___AttributeContext.Peer:
        """The destination of a network activity, such as accepting a TCP connection.
        In a multi hop network activity, the destination represents the receiver of
        the last hop.
        """

    @property
    def request(self) -> global___AttributeContext.Request:
        """Represents a network request, such as an HTTP request."""

    @property
    def response(self) -> global___AttributeContext.Response:
        """Represents a network response, such as an HTTP response."""

    @property
    def resource(self) -> global___AttributeContext.Resource:
        """Represents a target resource that is involved with a network activity.
        If multiple resources are involved with an activity, this must be the
        primary one.
        """

    @property
    def api(self) -> global___AttributeContext.Api:
        """Represents an API operation that is involved to a network activity."""

    def __init__(
        self,
        *,
        origin: global___AttributeContext.Peer | None = ...,
        source: global___AttributeContext.Peer | None = ...,
        destination: global___AttributeContext.Peer | None = ...,
        request: global___AttributeContext.Request | None = ...,
        response: global___AttributeContext.Response | None = ...,
        resource: global___AttributeContext.Resource | None = ...,
        api: global___AttributeContext.Api | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["api", b"api", "destination", b"destination", "origin", b"origin", "request", b"request", "resource", b"resource", "response", b"response", "source", b"source"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["api", b"api", "destination", b"destination", "origin", b"origin", "request", b"request", "resource", b"resource", "response", b"response", "source", b"source"]) -> None: ...

global___AttributeContext = AttributeContext



"Implements string conversion for `gnmi.Path`."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyright/strict and parsy don't mix well.
#   pyright: reportUnknownMemberType=false, reportUnknownArgumentType=false
#   pyright: reportUnknownParameterType=false, reportUnknownVariableType=false
#   pyright: reportFunctionMemberAccess=false, reportUnknownLambdaType=false

import re

import parsy as pa  # pyright: ignore[reportMissingTypeStubs]

from finsy.proto import gnmi

#################################
# Parser Combinator Definitions #
#################################

_SLASH = pa.string("/")
_LBRACK = pa.string("[")
_RBRACK = pa.string("]")
_EQUALS = pa.string("=")

__ID1 = pa.regex(r"[^\s\\/\[=\]]+").desc("_ID1")  # char+ no SPACE \ / [ ] =
# N.B. '[' is not allowed in a key unless it is escaped. This is required by
# unit tests, but this restriction could be removed in the future.
__KEY = pa.regex(r"[^\s\\\]=\[]+").desc("_KEY")  # char+ no SPACE \ [ ] =
__VAL = pa.regex(r"[^\\\]]+").desc("_VAL")  # char+ no \ ]
__ESC = pa.string("\\") >> (  # backslash escape sequence
    pa.char_from("\\/[]=")
    | pa.string("n").result("\n")
    | pa.string("r").result("\r")
    | pa.string("t").result("\t")
    | pa.regex(r"x[0-9a-fA-F]{2}").map(lambda s: chr(int(s[1:], 16)))
    | pa.regex(r"u[0-9a-fA-F]{4}").map(lambda s: chr(int(s[1:], 16)))
    | pa.regex(r"U[0-9a-fA-F]{8}").map(lambda s: chr(int(s[1:], 16)))
)

_IDENT = (__ID1 | __ESC).at_least(1).concat().desc("_IDENT")
_IDKEY = (__KEY | __ESC).at_least(1).concat().desc("_IDKEY")
_IDVAL = (__VAL | __ESC).at_least(1).concat().desc("_IDVAL")


@pa.generate
def _keyval():
    "Parse `[ IDKEY = IDVAL ]`"
    yield _LBRACK
    key = yield _IDKEY
    yield _EQUALS
    value = yield _IDVAL
    yield _RBRACK
    return key, value


@pa.generate
def _elem():
    "Parse `IDENT KEYVAL*`"
    ident = yield _IDENT
    kvs = yield _keyval.many()

    keys = {}
    for k, v in kvs:
        if k in keys:
            raise ValueError(f"parse failed: duplicate key {k!r}")
        keys[k] = v
    return ident, keys


@pa.generate
def _elems():
    """Parse zero or more `_elem` separated by `/`.
    Ignore a single slash at the beginning or end.
    """
    yield _SLASH.optional()
    elems = yield _elem.sep_by(_SLASH)
    yield _SLASH.optional()
    return elems


################################################################################

# This module provides two functions: `parse` and `to_str`.
#
# Example:
#
# ```
# path = gnmistring.parse("interfaces/interface[name=eth1]/state")
# ```
#
# This module does NOT allow the `origin` or `target` prefix properties to be
# specified in the string. Set these in `GNMIPath`.
#
# FIXME: This module does not support Unicode surrogate pairs.
#
# Reference:
# https://github.com/openconfig/reference/blob/master/rpc/gnmi/gnmi-path-strings.md


def parse(value: str) -> gnmi.Path:
    "Parse a string as a `gnmi.Path`."
    if value in ("/", ""):
        return gnmi.Path()

    try:
        elems = _elems.parse(value)
    except pa.ParseError as ex:
        # Map ParseError to ValueError for convenience.
        raise ValueError(f"parse failed: {ex} (value={value!r})") from ex

    result = gnmi.Path()
    for elem, key in elems:
        result.elem.add(name=elem, key=key)

    return result


def to_str(path: gnmi.Path) -> str:
    "Encode a `gnmi.Path` as a string."
    if not path.elem:
        return "/"

    return "/".join(_elem_str(elem) for elem in path.elem)


def _elem_str(elem: gnmi.PathElem) -> str:
    """Encode a `gnmi.PathElem` as a string."""
    name = _escape(elem.name, "[]/=", True)
    keyvals = [
        f"[{_escape(key,'[]=', True)}={_escape(val, ']', False)}]"
        for key, val in sorted(elem.key.items())
    ]

    return "".join([name, *keyvals])


_REPLACE_ESCAPES = re.compile(rb"\\x[0-9a-fA-F]{2}|\\t")


def _escape(value: str, chars: str, esc_space: bool) -> str:
    """Backslash escape the specified characters in value. If `esc_space` is
    True, we also escape spaces."""

    def _replace(m: re.Match[bytes]) -> bytes:
        s = m.group(0)
        if s[0:2] == rb"\t":
            return rb"\u0009"
        assert s[0:2] == rb"\x"
        return rb"\u00" + s[2:]

    # Canonical: Translate \xHH and \t escapes to \uHHHH.
    data = value.encode("unicode-escape")
    data = _REPLACE_ESCAPES.sub(_replace, data)

    table = {ord(char): f"\\{char}" for char in chars}
    if esc_space:  # space is escaped in ID and KEY (but not in VAL).
        table[0x20] = r"\u0020"

    return data.decode("ascii").translate(table)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: stratum1/p4_role_config.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dstratum1/p4_role_config.proto\x12\x07stratum\"\xe9\x01\n\x0cP4RoleConfig\x12\x18\n\x10\x65xclusive_p4_ids\x18\x01 \x03(\r\x12\x15\n\rshared_p4_ids\x18\x02 \x03(\r\x12<\n\x10packet_in_filter\x18\x03 \x01(\x0b\x32\".stratum.P4RoleConfig.PacketFilter\x12\x1b\n\x13receives_packet_ins\x18\x04 \x01(\x08\x12\x19\n\x11\x63\x61n_push_pipeline\x18\x05 \x01(\x08\x1a\x32\n\x0cPacketFilter\x12\x13\n\x0bmetadata_id\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x0c\x42\x03\x80\x01\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'stratum1.p4_role_config_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\200\001\000'
  _globals['_P4ROLECONFIG']._serialized_start=43
  _globals['_P4ROLECONFIG']._serialized_end=276
  _globals['_P4ROLECONFIG_PACKETFILTER']._serialized_start=226
  _globals['_P4ROLECONFIG_PACKETFILTER']._serialized_end=276
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2013-present Barefoot Networks, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class P4Data(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BITSTRING_FIELD_NUMBER: builtins.int
    VARBIT_FIELD_NUMBER: builtins.int
    BOOL_FIELD_NUMBER: builtins.int
    TUPLE_FIELD_NUMBER: builtins.int
    STRUCT_FIELD_NUMBER: builtins.int
    HEADER_FIELD_NUMBER: builtins.int
    HEADER_UNION_FIELD_NUMBER: builtins.int
    HEADER_STACK_FIELD_NUMBER: builtins.int
    HEADER_UNION_STACK_FIELD_NUMBER: builtins.int
    ENUM_FIELD_NUMBER: builtins.int
    ERROR_FIELD_NUMBER: builtins.int
    ENUM_VALUE_FIELD_NUMBER: builtins.int
    bitstring: builtins.bytes
    """for bit<W>, int<W>"""
    bool: builtins.bool
    enum: builtins.str
    """safe (non-serializable) enums only"""
    error: builtins.str
    enum_value: builtins.bytes
    """serializable enums only"""
    @property
    def varbit(self) -> global___P4Varbit:
        """for varbit<W>"""

    @property
    def tuple(self) -> global___P4StructLike: ...
    @property
    def struct(self) -> global___P4StructLike: ...
    @property
    def header(self) -> global___P4Header: ...
    @property
    def header_union(self) -> global___P4HeaderUnion: ...
    @property
    def header_stack(self) -> global___P4HeaderStack: ...
    @property
    def header_union_stack(self) -> global___P4HeaderUnionStack: ...
    def __init__(
        self,
        *,
        bitstring: builtins.bytes = ...,
        varbit: global___P4Varbit | None = ...,
        bool: builtins.bool = ...,
        tuple: global___P4StructLike | None = ...,
        struct: global___P4StructLike | None = ...,
        header: global___P4Header | None = ...,
        header_union: global___P4HeaderUnion | None = ...,
        header_stack: global___P4HeaderStack | None = ...,
        header_union_stack: global___P4HeaderUnionStack | None = ...,
        enum: builtins.str = ...,
        error: builtins.str = ...,
        enum_value: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bitstring", b"bitstring", "bool", b"bool", "data", b"data", "enum", b"enum", "enum_value", b"enum_value", "error", b"error", "header", b"header", "header_stack", b"header_stack", "header_union", b"header_union", "header_union_stack", b"header_union_stack", "struct", b"struct", "tuple", b"tuple", "varbit", b"varbit"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bitstring", b"bitstring", "bool", b"bool", "data", b"data", "enum", b"enum", "enum_value", b"enum_value", "error", b"error", "header", b"header", "header_stack", b"header_stack", "header_union", b"header_union", "header_union_stack", b"header_union_stack", "struct", b"struct", "tuple", b"tuple", "varbit", b"varbit"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["data", b"data"]) -> typing.Literal["bitstring", "varbit", "bool", "tuple", "struct", "header", "header_union", "header_stack", "header_union_stack", "enum", "error", "enum_value"] | None: ...

global___P4Data = P4Data

@typing.final
class P4Varbit(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BITSTRING_FIELD_NUMBER: builtins.int
    BITWIDTH_FIELD_NUMBER: builtins.int
    bitstring: builtins.bytes
    bitwidth: builtins.int
    """dynamic bitwidth of the field"""
    def __init__(
        self,
        *,
        bitstring: builtins.bytes = ...,
        bitwidth: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["bitstring", b"bitstring", "bitwidth", b"bitwidth"]) -> None: ...

global___P4Varbit = P4Varbit

@typing.final
class P4StructLike(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MEMBERS_FIELD_NUMBER: builtins.int
    @property
    def members(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4Data]: ...
    def __init__(
        self,
        *,
        members: collections.abc.Iterable[global___P4Data] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["members", b"members"]) -> None: ...

global___P4StructLike = P4StructLike

@typing.final
class P4Header(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IS_VALID_FIELD_NUMBER: builtins.int
    BITSTRINGS_FIELD_NUMBER: builtins.int
    is_valid: builtins.bool
    """If the header is invalid (is_valid is "false"), then the bitstrings
    repeated field must be empty.
    """
    @property
    def bitstrings(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(
        self,
        *,
        is_valid: builtins.bool = ...,
        bitstrings: collections.abc.Iterable[builtins.bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["bitstrings", b"bitstrings", "is_valid", b"is_valid"]) -> None: ...

global___P4Header = P4Header

@typing.final
class P4HeaderUnion(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALID_HEADER_NAME_FIELD_NUMBER: builtins.int
    VALID_HEADER_FIELD_NUMBER: builtins.int
    valid_header_name: builtins.str
    """An empty string indicates that none of the union members are valid and
    valid_header must therefore be unset.
    """
    @property
    def valid_header(self) -> global___P4Header: ...
    def __init__(
        self,
        *,
        valid_header_name: builtins.str = ...,
        valid_header: global___P4Header | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["valid_header", b"valid_header"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["valid_header", b"valid_header", "valid_header_name", b"valid_header_name"]) -> None: ...

global___P4HeaderUnion = P4HeaderUnion

@typing.final
class P4HeaderStack(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENTRIES_FIELD_NUMBER: builtins.int
    @property
    def entries(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4Header]:
        """The length of this repeated field must always be equal to the compile-time
        size of the header stack, which is specified in P4Info.
        """

    def __init__(
        self,
        *,
        entries: collections.abc.Iterable[global___P4Header] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["entries", b"entries"]) -> None: ...

global___P4HeaderStack = P4HeaderStack

@typing.final
class P4HeaderUnionStack(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENTRIES_FIELD_NUMBER: builtins.int
    @property
    def entries(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4HeaderUnion]:
        """The length of this repeated field must always be equal to the compile-time
        size of the header union stack, which is specified in P4Info.
        """

    def __init__(
        self,
        *,
        entries: collections.abc.Iterable[global___P4HeaderUnion] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["entries", b"entries"]) -> None: ...

global___P4HeaderUnionStack = P4HeaderUnionStack



"Implements the Arbitrator class."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyright: reportPrivateUsage=false

import finsy.switch as _sw  # break circular import
from finsy import pbutil
from finsy.grpcutil import GRPCStatusCode
from finsy.log import LOGGER
from finsy.p4client import P4ClientError, P4RpcStatus
from finsy.proto import U128, p4r

_NOT_ASSIGNED = -1


class Arbitrator:
    """Manages state used for client/role arbitration.

    TODO: Eventually, make an AbtractArbitrator base class to handle
    the client arbitration, so different algorithms can be plugged in.
    """

    initial_election_id: int
    election_id: int
    role: p4r.Role | None
    is_primary: bool = False
    primary_id: int = _NOT_ASSIGNED

    def __init__(
        self,
        initial_election_id: int,
        role_name: str = "",
        role_config: pbutil.PBMessage | None = None,
    ):
        self.initial_election_id = initial_election_id
        self.election_id = initial_election_id
        self.role = _create_role(role_name, role_config)

    @property
    def role_name(self) -> str:
        "Role name or '' if there is no role set."
        if self.role is None:
            return ""
        return self.role.name

    async def handshake(self, switch: "_sw.Switch", *, conflict: bool = False) -> None:
        """Perform the P4Runtime client arbitration handshake."""
        assert not self.is_primary

        if conflict:
            # Current election_id conflicts with another connection, so start
            # the bidding at one less.
            self.election_id -= 1

        response = await self._arbitration_request(switch)
        status = P4RpcStatus.from_status(response.status)
        primary_id = U128.decode(response.election_id)

        while status.code == GRPCStatusCode.NOT_FOUND:
            self.election_id = primary_id

            response = await self._arbitration_request(switch)
            status = P4RpcStatus.from_status(response.status)
            primary_id = U128.decode(response.election_id)

        assert status.code in (GRPCStatusCode.OK, GRPCStatusCode.ALREADY_EXISTS)

        self.primary_id = primary_id
        self.is_primary = status.code == GRPCStatusCode.OK
        self._check_invariant()

    async def update(
        self,
        switch: "_sw.Switch",
        msg: p4r.MasterArbitrationUpdate,
    ) -> None:
        "Called with subsequent arbitration update responses."
        status_code = P4RpcStatus.from_status(msg.status).code
        new_primary_id = U128.decode(msg.election_id)

        if new_primary_id >= self.primary_id:
            self.primary_id = new_primary_id
        else:
            # TECHDEBT: simple_switch_grpc sends a status code of OK and a
            # decreased election_id when it wants the next backup client to
            # become the primary.
            LOGGER.warning("election_id decreased to %r", new_primary_id)
            if (
                status_code == GRPCStatusCode.OK
                and not self.is_primary
                and new_primary_id == self.election_id
            ):
                status_code = GRPCStatusCode.NOT_FOUND

        match status_code:
            case GRPCStatusCode.OK:
                if not self.is_primary:
                    self.is_primary = True
                    switch._become_primary()
            case GRPCStatusCode.ALREADY_EXISTS:
                if self.is_primary:
                    self.is_primary = False
                    switch._become_backup()
            case GRPCStatusCode.NOT_FOUND:
                self.is_primary = False
                await self._request_primary(switch)
            case other:
                raise ValueError(f"Unexpected status: {other!r}")

    def reset(self) -> None:
        "Called when client stream disconnects."
        self.election_id = self.initial_election_id
        self.is_primary = False
        self.primary_id = _NOT_ASSIGNED

    def complete_request(self, msg: pbutil.PBMessage) -> None:
        "Complete request with role/election_id information."
        if isinstance(msg, p4r.ReadRequest):
            if self.role is not None:
                msg.role = self.role.name

        elif isinstance(
            msg,
            (p4r.SetForwardingPipelineConfigRequest, p4r.WriteRequest),
        ):
            if self.role is not None:
                msg.role = self.role.name

            msg.election_id.CopyFrom(U128.encode(self.election_id))

    async def _arbitration_request(self, switch: "_sw.Switch"):
        "Send a MasterArbitrationUpdate request and wait for the response."
        for _ in range(5):
            await self._send(switch)

            response = await self._receive(switch)
            if response is not None:
                return response

            self.election_id -= 1

        raise ValueError("no compatible election_id")

    async def _request_primary(self, switch: "_sw.Switch"):
        "Send a request to become the new primary."
        self.election_id = self.primary_id
        await self._send(switch)

    async def _send(self, switch: "_sw.Switch") -> None:
        "Send MasterArbitrationUpdate message to switch."
        assert self.election_id >= 0
        assert switch._p4client is not None

        request = p4r.StreamMessageRequest(
            arbitration=p4r.MasterArbitrationUpdate(
                role=self.role,
                device_id=switch.device_id,
                election_id=U128.encode(self.election_id),
            )
        )

        await switch._p4client.send(request)

    async def _receive(
        self,
        switch: "_sw.Switch",
    ) -> p4r.MasterArbitrationUpdate | None:
        """Wait for MasterArbitrationUpdate response from switch.

        Return None if the response indicates the `election_id` is in use.

        Log and ignore any other responses from the switch. Since this method
        can wait indefinitely, the caller should put a timeout on this method.
        """
        assert switch._p4client is not None

        try:
            while True:
                response = await switch._p4client.receive()
                if response.WhichOneof("update") == "arbitration":
                    break
                LOGGER.warning("Arbitrator ignored response: %r", response)

        except P4ClientError as ex:
            if ex.is_election_id_used:
                return None
            raise

        # TODO: Match arbitration response to request we sent?

        return response.arbitration

    def _check_invariant(self):
        "Check the Arbitrator's invariant."
        if self.is_primary:
            # When we're the primary, election_id must equal primary_id.
            if self.election_id != self.primary_id:
                raise RuntimeError(
                    f"primary invariant failed: election_id={self.election_id}, primary_id={self.primary_id}"
                )
        else:
            # When we're the backup, election_id must be less than primary_id.
            # FIXME: Unless we're renegotiating?
            if self.election_id >= self.primary_id:
                raise RuntimeError(
                    f"backup invariant failed: election_id={self.election_id}, primary_id={self.primary_id}"
                )


def _create_role(
    role_name: str,
    role_config: pbutil.PBMessage | None,
) -> p4r.Role | None:
    "Create a new P4Runtime Role object."
    if not role_name and role_config is None:
        return None

    if role_config is None:
        raise ValueError("role_config cannot be None when role_name is set")

    return p4r.Role(
        name=role_name,
        # [2021-10-18] If I remove the `pbutil.to_any()` and try
        # to set the message field `config=role_config`, I get a segmentation
        # fault on MacOS. [protobuf 4.21.7]
        config=pbutil.to_any(role_config),
    )



"Implements the MACAddress class."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

_BYTE_WIDTH = 6
_BIT_WIDTH = _BYTE_WIDTH * 8


@functools.total_ordering
class MACAddress:
    """Concrete class for a MAC address."""

    __slots__ = ("_mac", "__weakref__")
    _mac: int

    def __init__(self, value: object) -> None:
        "Construct a MAC address from a string, integer or bytes object."
        match value:
            case int():
                self._mac = _from_int(value)
            case bytes():
                self._mac = _from_bytes(value)
            case MACAddress():
                self._mac = value._mac
            case _:
                # Input argument is a MAC string, or an object that is
                # formatted as MAC string (behave like ipaddress module).
                self._mac = _from_string(str(value))

    @property
    def packed(self) -> bytes:
        "Return the MAC address a byte string."
        return _to_bytes(self._mac)

    @property
    def max_prefixlen(self) -> int:
        "Return the maximum prefix length (48 bits)."
        return _BIT_WIDTH

    @property
    def is_multicast(self) -> bool:
        "Return true if MAC address has the multicast bit set."
        return self._mac & (1 << 40) != 0

    @property
    def is_private(self) -> bool:
        "Return true if MAC address has the locally administered bit set."
        return self._mac & (1 << 41) != 0

    @property
    def is_global(self) -> bool:
        "Return true if the locally administered bit is not set."
        return not self.is_private

    @property
    def is_unspecified(self) -> bool:
        "Return true if MAC address is all zeros."
        return self._mac == 0

    @property
    def is_broadcast(self) -> bool:
        "Return true if MAC address is the broadcast address."
        return self._mac == (1 << _BIT_WIDTH) - 1

    def __int__(self) -> int:
        return self._mac

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, MACAddress):
            return NotImplemented
        return self._mac == rhs._mac

    def __lt__(self, rhs: object) -> bool:
        if not isinstance(rhs, MACAddress):
            return NotImplemented
        return self._mac < rhs._mac

    def __repr__(self) -> str:
        return f"MACAddress({_to_string(self._mac)!r})"

    def __str__(self) -> str:
        return _to_string(self._mac)

    def __hash__(self) -> int:
        return hash(hex(self._mac))


def _from_string(value: str) -> int:
    """Parse a MAC address string and return an integer representation.

    Args:
        value: A string of the form "hh:hh:hh:hh:hh:hh" or "hh-hh-hh-hh-hh-hh".
          where `hh` is 1-2 hexadecimal digits.

    Returns:
        The MAC address as an integer.

    Raises:
        ValueError: String value is invalid.
    """
    try:
        delimiter = "-" if "-" in value else ":"
        data = bytes(_parse_octet(octet) for octet in value.split(delimiter))
        return _from_bytes(data)
    except ValueError:
        pass

    raise ValueError(f"invalid MAC address: {value!r}")


def _parse_octet(value: str) -> int:
    "Parse a 2-char octet in string format."
    length = len(value)
    if length == 1 or length == 2 == len(value.strip()):
        # Be careful because int() will accept/ignore space chars.
        return int(value, 16)
    raise ValueError("invalid octet")


def _from_bytes(value: bytes) -> int:
    "Convert 6-bytes to integer."
    if len(value) != _BYTE_WIDTH:
        raise ValueError(f"invalid MAC address: {value!r}")
    return int.from_bytes(value, byteorder="big")


def _from_int(value: int) -> int:
    "Check that integer value for MAC address is in range."
    if value < 0 or value > (1 << _BIT_WIDTH) - 1:
        raise ValueError(f"invalid MAC address: {value!r}")
    return value


def _to_string(value: int) -> str:
    "Convert integer MAC address value to string."
    return ":".join(f"{octet:02x}" for octet in _to_bytes(value))


def _to_bytes(value: int) -> bytes:
    "Convert integer MAC address value to bytes."
    return value.to_bytes(_BYTE_WIDTH, byteorder="big")



"Implements entity encode/decode functions."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import collections.abc
from dataclasses import KW_ONLY, dataclass
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    NoReturn,
    Protocol,
    Sequence,
    TypeVar,
)

from typing_extensions import Self

from finsy import p4values, pbutil
from finsy.log import LOGGER
from finsy.p4schema import (
    P4Action,
    P4ActionRef,
    P4Schema,
    P4Table,
    P4UpdateType,
    P4ValueSet,
)
from finsy.proto import p4r

NOACTION_STR = "NoAction()"


class _SupportsDecodeEntity(Protocol):
    @classmethod
    def decode(
        cls,
        msg: Any,  # (p4r.Entity | p4r.StreamMessageResponse)
        schema: P4Schema,
    ) -> Self:
        "Decode message to produce entity or stream response."
        ...  # pragma: no cover


class _SupportsEncodeEntity(Protocol):
    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode object as an entity."
        ...  # pragma: no cover


class _SupportsEncodeUpdate(Protocol):
    def encode_update(self, schema: P4Schema) -> p4r.Update | p4r.StreamMessageRequest:
        "Encode object as an update or stream request."
        ...  # pragma: no cover


_DECODER: dict[str, _SupportsDecodeEntity] = {}

_D = TypeVar("_D", bound=_SupportsDecodeEntity)


def decodable(key: str) -> Callable[[_D], _D]:
    "Class decorator to specify the class used to decode P4Runtime messages."

    def _decorate(cls: _D) -> _D:
        assert key not in _DECODER
        _DECODER[key] = cls
        return cls

    return _decorate


def decode_entity(msg: p4r.Entity, schema: P4Schema) -> Any:
    "Decode a P4Runtime Entity to the Python class registered in _DECODER."
    key = msg.WhichOneof("entity")
    if key is None:
        raise ValueError("missing entity")

    if key == "packet_replication_engine_entry":
        sub_msg = msg.packet_replication_engine_entry
        key = sub_msg.WhichOneof("type")
        if key is None:
            raise ValueError("missing packet_replication_engine type")

    return _DECODER[key].decode(msg, schema)


def decode_stream(msg: p4r.StreamMessageResponse, schema: P4Schema) -> Any:
    "Decode a StreamMessageResponse to the class registered in _DECODER."
    key = msg.WhichOneof("update")
    if key is None:
        raise ValueError("missing update")

    return _DECODER[key].decode(msg, schema)


# Recursive type vars.
P4EntityList = p4r.Entity | _SupportsEncodeEntity | Iterable["P4EntityList"]
P4UpdateList = (
    p4r.Update
    | p4r.StreamMessageRequest
    | _SupportsEncodeUpdate
    | Iterable["P4UpdateList"]
)


def flatten(values: Any) -> Iterator[Any]:
    "Flatten lists and tuples."
    for val in values:
        if isinstance(val, collections.abc.Iterable):
            yield from flatten(val)
        else:
            yield val


def _encode_entity(
    value: p4r.Entity | _SupportsEncodeEntity,
    schema: P4Schema,
) -> p4r.Entity:
    "Encode an entity, if necessary."
    if isinstance(value, p4r.Entity):
        return value

    return value.encode(schema)


def encode_entities(
    values: Iterable[P4EntityList], schema: P4Schema
) -> list[p4r.Entity]:
    """Convert list of python objects to list of P4Runtime Entities."""
    return [_encode_entity(val, schema) for val in flatten(values)]


def _encode_update(
    value: p4r.Update | p4r.StreamMessageRequest | _SupportsEncodeUpdate,
    schema: P4Schema,
) -> p4r.Update | p4r.StreamMessageRequest:
    "Encode an Update or outgoing message request, if necessary."
    if isinstance(value, (p4r.Update, p4r.StreamMessageRequest)):
        return value

    return value.encode_update(schema)


def encode_updates(
    values: P4UpdateList,
    schema: P4Schema,
) -> list[p4r.Update | p4r.StreamMessageRequest]:
    """Convert list of python objects to P4Runtime Updates or request messages."""
    if not isinstance(values, collections.abc.Iterable):
        return [_encode_update(values, schema)]

    return [_encode_update(val, schema) for val in flatten(values)]


class P4Entity(abc.ABC):
    """Abstract base class for P4Entity subclasses.

    All entities have `encode` and `decode` methods.

    The `encode` method is used to encode the entity to a Protobuf message. The
    `decode` method is a class method that decodes a Protobuf message and
    returns the entity object.
    """

    @abc.abstractmethod
    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode object as an entity."

    @classmethod
    @abc.abstractmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode Protobuf to entity object."


class _P4Writable(P4Entity):
    """Abstract base class for entities that support insert/modify/delete.

    A writable entity can specify the type of update by using a unary operator.

    1. `+` means INSERT.
    2. `-` means DELETE.
    3. `~` means MODIFY.

    Note: For efficiency, this class just mutates itself! This is safe as long
    as entities are used only once.
    """

    _update_type: P4UpdateType = P4UpdateType.UNSPECIFIED

    def __pos__(self) -> Self:
        if self._update_type != P4UpdateType.UNSPECIFIED:
            raise ValueError("update type already specified")
        self._update_type = P4UpdateType.INSERT
        return self

    def __neg__(self) -> Self:
        if self._update_type != P4UpdateType.UNSPECIFIED:
            raise ValueError("update type already specified")
        self._update_type = P4UpdateType.DELETE
        return self

    def __invert__(self) -> Self:
        if self._update_type != P4UpdateType.UNSPECIFIED:
            raise ValueError("update type already specified")
        self._update_type = P4UpdateType.MODIFY
        return self

    def encode_update(self, schema: P4Schema) -> p4r.Update:
        "Encode object as a Protobuf Update message."
        if self._update_type == P4UpdateType.UNSPECIFIED:
            raise ValueError(f"unspecified update type (+, ~, -): {self!r}")
        return p4r.Update(type=self._update_type.vt(), entity=self.encode(schema))


class _P4ModifyOnly(P4Entity):
    """Abstract base class for entities that only support modify.

    You can use `~` to indicate MODIFY, but this is optional.
    """

    def __invert__(self) -> Self:
        return self

    def encode_update(self, schema: P4Schema) -> p4r.Update:
        "Encode object as a Protobuf Update message."
        return p4r.Update(type=P4UpdateType.MODIFY.vt(), entity=self.encode(schema))


_DataValueType = Any  # TODO: tighten P4Data type constraint later
_MetadataDictType = dict[str, Any]
_PortType = int
_ReplicaCanonType = tuple[_PortType, int]
_ReplicaType = _ReplicaCanonType | _PortType

P4Weight = int | tuple[int, int]
P4WeightedAction = tuple[P4Weight, "P4TableAction"]


def encode_replica(value: _ReplicaType) -> p4r.Replica:
    "Convert python representation to Replica."
    if isinstance(value, int):
        egress_port, instance = value, 0
    else:
        assert isinstance(value, tuple) and len(value) == 2
        egress_port, instance = value

    return p4r.Replica(egress_port=egress_port, instance=instance)


def decode_replica(replica: p4r.Replica) -> _ReplicaCanonType:
    "Convert Replica to python representation."
    return (replica.egress_port, replica.instance)


def format_replica(value: _ReplicaType) -> str:
    "Format python representation of Replica."
    if isinstance(value, int):
        return str(value)
    assert isinstance(value, tuple) and len(value) == 2
    port, instance = value
    if instance == 0:
        return str(port)
    return f"{port}#{instance}"


def encode_watch_port(watch_port: int) -> bytes:
    "Encode watch_port into protobuf message."
    return p4values.encode_exact(watch_port, 32)


def decode_watch_port(watch_port: bytes) -> int:
    "Decode watch_port from protobuf message."
    return int(p4values.decode_exact(watch_port, 32))


class P4TableMatch(dict[str, Any]):
    """Represents a set of P4Runtime field matches.

    Each match field is stored as a dictionary key, where the key is the name
    of the match field. The field's value should be appropriate for the type
    of match (EXACT, LPM, TERNARY, etc.)

    Construct a match similar to a dictionary.

    Example:
    ```
    # Keyword arguments:
    match = P4TableMatch(ipv4_dst="10.0.0.1")

    # Dictionary argument:
    match = P4TableMatch({"ipv4_dst": "10.0.0.1"})

    # List of 2-tuples:
    match = P4TableMatch([("ipv4_dst", "10.0.0.1")])
    ```

    P4TableMatch is implemented as a subclass of `dict`. It supports all of the
    standard dictionary methods:
    ```
    match = P4TableMatch()
    match["ipv4_dst"] = "10.0.0.1"
    assert len(match) == 1
    ```

    Reference "9.1.1 Match Format":
        Each match field is translated to a FieldMatch Protobuf message by
        translating the entry key to a `field_id`. The type of match (EXACT,
        LPM, TERNARY, OPTIONAL, or RANGE) is determined by the P4Info, and the
        value is converted to the Protobuf representation.

    Supported String Values:
    ```
    EXACT: "255", "0xFF", "10.0.0.1", "2000::1", "00:00:00:00:00:01"

    LPM: "255/8", "0xFF/8", "10.0.0.1/32", "2000::1/128",
            "00:00:00:00:00:01/48"
        (+ all exact formats are promoted to all-1 masks)

    TERNARY: "255/&255", "0xFF/&0xFF", "10.0.0.1/&255.255.255.255",
        "2000::1/&128", "00:00:00:00:00:01/&48"
        (+ all exact formats are promoted to all-1 masks)
        (+ all lpm formats are promoted to the specified contiguous mask)

    RANGE: "0...255", "0x00...0xFF", "10.0.0.1...10.0.0.9",
        "2000::1...2001::9", "00:00:00:00:00:01...00:00:00:00:00:09"
        (+ all exact formats are promoted to single-value ranges)

    OPTIONAL: Same as exact format.
    ```

    See the `p4values.py` module for all supported value classes.

    TODO:
        - Change range delimiter to '-' (and drop '-' delimited MAC's).
        - Consider supporting ternary values with just '/' (and drop support
          for decimal masks; mask must be hexadecimal number).

    See Also:
        - P4TableEntry
    """

    def encode(self, table: P4Table) -> list[p4r.FieldMatch]:
        "Encode TableMatch data as a list of Protobuf fields."
        result: list[p4r.FieldMatch] = []
        match_fields = table.match_fields

        for key, value in self.items():
            try:
                field = match_fields[key].encode_field(value)
                if field is not None:
                    result.append(field)
            except Exception as ex:
                raise ValueError(f"{table.name!r}: Match field {key!r}: {ex}") from ex

        return result

    @classmethod
    def decode(cls, msgs: Iterable[p4r.FieldMatch], table: P4Table) -> Self:
        "Decode Protobuf fields as TableMatch data."
        result: dict[str, Any] = {}
        match_fields = table.match_fields

        for field in msgs:
            fld = match_fields[field.field_id]
            result[fld.alias] = fld.decode_field(field)

        return cls(result)

    def format_dict(
        self,
        table: P4Table,
        *,
        wildcard: str | None = None,
    ) -> dict[str, str]:
        """Format the table match fields as a human-readable dictionary.

        The result is a dictionary showing the TableMatch data for fields
        included in the match. If `wildcard` is specified, all fields defined
        in P4Info will be included with their value set to the wildcard string.

        Values are formatted using the format/type specified in P4Info.
        """
        result: dict[str, str] = {}

        for fld in table.match_fields:
            value = self.get(fld.alias, None)
            if value is not None:
                result[fld.alias] = fld.format_field(value)
            elif wildcard is not None:
                result[fld.alias] = wildcard

        return result

    def format_str(
        self,
        table: P4Table,
        *,
        wildcard: str | None = None,
    ) -> str:
        """Format the table match fields as a human-readable string.

        The result is a string showing the TableMatch data for fields included
        in the match. If `wildcard` is specified, all fields defined in P4Info
        will be included with their value set to the wildcard string.

        All fields are formatted as "name=value" and they are delimited by
        spaces.

        Values are formatted using the format/type specified in P4Info.
        """
        result: list[str] = []

        for fld in table.match_fields:
            value = self.get(fld.alias, None)
            if value is not None:
                result.append(f"{fld.alias}={fld.format_field(value)}")
            elif wildcard is not None:
                result.append(f"{fld.alias}={wildcard}")

        return " ".join(result)


@dataclass(init=False, slots=True)
class P4TableAction:
    """Represents a P4Runtime Action reference for a direct table.

    Attributes:
        name (str): the name of the action.
        args (dict[str, Any]): the action's arguments as a dictionary.

    Example:
        If the name of the action is "ipv4_forward" and it takes a single
        "port" parameter, you can construct the action as:

        ```
        action = P4TableAction("ipv4_forward", port=1)
        ```

    Reference "9.1.2 Action Specification":
        The Action Protobuf has fields: (action_id, params). Finsy translates
        `name` to the appropriate `action_id` as determined by P4Info. It also
        translates each named argument in `args` to the appropriate `param_id`.

    See Also:
        To specify an action for an indirect table, use `P4IndirectAction`.
        Note that P4TableAction will automatically be promoted to an "indirect"
        action if needed.

    Operators:
        A `P4TableAction` supports the multiplication operator (*) for
        constructing "weighted actions". A weighted action is used in specifying
        indirect actions. Here is an action with a weight of 3:

        ```
        weighted_action = 3 * P4TableAction("ipv4_forward", port=1)
        ```

        To specify a weight with a `watch_port`, use a tuple `(weight, port)`.
        The weight is always a positive integer.

    See Also:
        - P4TableEntry
    """

    name: str
    "The name of the action."
    args: dict[str, Any]
    "The action's arguments as a dictionary."

    def __init__(self, __name: str, /, **args: Any):
        self.name = __name
        self.args = args

    def encode_table_action(self, table: P4Table) -> p4r.TableAction:
        """Encode TableAction data as protobuf.

        If the table is indirect, promote the action to a "one-shot" indirect
        action.
        """
        try:
            action = table.actions[self.name]
        except Exception as ex:
            raise ValueError(f"{table.name!r}: {ex}") from ex

        action_p4 = self._encode_action(action)

        if table.action_profile is not None:
            # Promote action to ActionProfileActionSet entry with weight=1.
            return p4r.TableAction(
                action_profile_action_set=p4r.ActionProfileActionSet(
                    action_profile_actions=[
                        p4r.ActionProfileAction(action=action_p4, weight=1)
                    ]
                )
            )

        return p4r.TableAction(action=action_p4)

    def _fail_missing_params(self, action: P4ActionRef | P4Action) -> NoReturn:
        "Report missing parameters."
        seen = {param.name for param in action.params}
        for name in self.args:
            param = action.params[name]
            seen.remove(param.name)

        raise ValueError(f"Action {action.alias!r}: missing parameters {seen}")

    def encode_action(self, schema: P4Schema | P4Table) -> p4r.Action:
        "Encode Action data as protobuf."
        action = schema.actions[self.name]
        return self._encode_action(action)

    def _encode_action(self, action: P4ActionRef | P4Action) -> p4r.Action:
        "Helper to encode an action."
        aps = action.params
        try:
            params = [
                aps[name].encode_param(value) for name, value in self.args.items()
            ]
        except ValueError as ex:
            raise ValueError(f"{action.alias!r}: {ex}") from ex

        # Check for missing action parameters. We always accept an action with
        # no parameters (for wildcard ReadRequests).
        param_count = len(params)
        if param_count > 0 and param_count != len(aps):
            self._fail_missing_params(action)

        return p4r.Action(action_id=action.id, params=params)

    @classmethod
    def decode_table_action(
        cls, msg: p4r.TableAction, table: P4Table
    ) -> Self | "P4IndirectAction":
        "Decode protobuf to TableAction data."
        match msg.WhichOneof("type"):
            case "action":
                return cls.decode_action(msg.action, table)
            case "action_profile_member_id":
                return P4IndirectAction(member_id=msg.action_profile_member_id)
            case "action_profile_group_id":
                return P4IndirectAction(group_id=msg.action_profile_group_id)
            case "action_profile_action_set":
                return P4IndirectAction.decode_action_set(
                    msg.action_profile_action_set, table
                )
            case other:
                raise ValueError(f"unknown oneof: {other!r}")

    @classmethod
    def decode_action(cls, msg: p4r.Action, parent: P4Schema | P4Table) -> Self:
        "Decode protobuf to Action data."
        action = parent.actions[msg.action_id]
        args = {}
        for param in msg.params:
            action_param = action.params[param.param_id]
            value = action_param.decode_param(param)
            args[action_param.name] = value

        return cls(action.alias, **args)

    def format_str(self, schema: P4Schema | P4Table) -> str:
        """Format the table action as a human-readable string.

        The result is formatted to look like a function call:

        ```
        name(param1=value1, ...)
        ```

        Where `name` is the action name, and `(param<N>, value<N>)` are the
        action parameters. The format of `value<N>` is schema-dependent.
        """
        aps = schema.actions[self.name].params
        args = [
            f"{key}={aps[key].format_param(value)}" for key, value in self.args.items()
        ]

        return f"{self.name}({', '.join(args)})"

    def __mul__(self, weight: P4Weight) -> P4WeightedAction:
        "Make a weighted action."
        if not isinstance(
            weight, (int, tuple)
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise NotImplementedError("expected P4Weight")
        return (weight, self)

    def __rmul__(self, weight: P4Weight) -> P4WeightedAction:
        "Make a weighted action."
        if not isinstance(
            weight, (int, tuple)
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise NotImplementedError("expected P4Weight")
        return (weight, self)

    def __call__(self, **params: Any) -> Self:
        "Return a new action with the updated parameters."
        return self.__class__(self.name, **(self.args | params))


@dataclass(slots=True)
class P4IndirectAction:
    """Represents a P4Runtime Action reference for an indirect table.

    An indirect action can be either:

    1. a "one-shot" action (action_set)
    2. a reference to an action profile member (member_id)
    3. a reference to an action profile group (group_id)

    Only one of action_set, member_id or group_id may be configured. The other
    values must be None.

    Attributes:
        action_set: sequence of weighted actions for one-shot
        member_id: ID of action profile member
        group_id: ID of action profile group

    Examples:

    ```python
    # Construct a one-shot action profile.
    one_shot = P4IndirectAction(
        2 * P4TableAction("forward", port=1),
        1 * P4TableAction("forward", port=2),
    )

    # Refer to an action profile member by ID.
    member_action = P4IndirectAction(member_id=1)

    # Refer to an action profile group by ID.
    group_action = P4IndirectAction(group_id=2)
    ```

    References:
        - "9.1.2. Action Specification",
        - "9.2.3. One Shot Action Selector Programming"

    TODO: Refactor into three classes? P4OneShotAction, P4MemberAction, and
        P4GroupAction.

    See Also:
        - P4TableEntry
    """

    action_set: Sequence[P4WeightedAction] | None = None
    "Sequence of weighted actions defining one-shot action profile."
    _: KW_ONLY
    member_id: int | None = None
    "ID of action profile member."
    group_id: int | None = None
    "ID of action profile group."

    def __post_init__(self) -> None:
        if not self._check_invariant():
            raise ValueError(
                "exactly one of action_set, member_id, or group_id must be set"
            )

    def _check_invariant(self) -> bool:
        "Return true if instance satisfies class invariant."
        if self.action_set is not None:
            return self.member_id is None and self.group_id is None
        if self.member_id is not None:
            return self.group_id is None
        return self.group_id is not None

    def encode_table_action(self, table: P4Table) -> p4r.TableAction:
        "Encode object as a TableAction."
        if self.action_set is not None:
            return p4r.TableAction(
                action_profile_action_set=self.encode_action_set(table)
            )

        if self.member_id is not None:
            return p4r.TableAction(action_profile_member_id=self.member_id)

        assert self.group_id is not None
        return p4r.TableAction(action_profile_group_id=self.group_id)

    def encode_action_set(self, table: P4Table) -> p4r.ActionProfileActionSet:
        "Encode object as an ActionProfileActionSet."
        assert self.action_set is not None

        profile_actions: list[p4r.ActionProfileAction] = []
        for weight, table_action in self.action_set:
            action = table_action.encode_action(table)

            match weight:
                case int(weight_value):
                    watch_port = None
                case (weight_value, int(watch)):
                    watch_port = encode_watch_port(watch)
                case _:  # pyright: ignore[reportUnnecessaryComparison]
                    raise ValueError(f"unexpected action weight: {weight!r}")

            profile = p4r.ActionProfileAction(action=action, weight=weight_value)
            if watch_port is not None:
                profile.watch_port = watch_port
            profile_actions.append(profile)

        return p4r.ActionProfileActionSet(action_profile_actions=profile_actions)

    @classmethod
    def decode_action_set(cls, msg: p4r.ActionProfileActionSet, table: P4Table) -> Self:
        "Decode ActionProfileActionSet."
        action_set = list[P4WeightedAction]()

        for action in msg.action_profile_actions:
            match action.WhichOneof("watch_kind"):
                case "watch_port":
                    weight = (action.weight, decode_watch_port(action.watch_port))
                case None:
                    weight = action.weight
                case other:
                    # "watch" (deprecated) is not supported
                    raise ValueError(f"unexpected oneof: {other!r}")

            table_action = P4TableAction.decode_action(action.action, table)
            action_set.append((weight, table_action))

        return cls(action_set)

    def format_str(self, table: P4Table) -> str:
        """Format the indirect table action as a human-readable string."""
        if self.action_set is not None:
            weighted_actions = [
                f"{weight}*{action.format_str(table)}"
                for weight, action in self.action_set
            ]
            return " ".join(weighted_actions)

        # Use the name of the action_profile, if we can get it. If not, just
        # use the placeholder "__indirect".
        if table.action_profile is not None:
            profile_name = f"@{table.action_profile.alias}"
        else:
            profile_name = "__indirect"

        if self.member_id is not None:
            return f"{profile_name}[[{self.member_id:#x}]]"

        return f"{profile_name}[{self.group_id:#x}]"

    def __repr__(self) -> str:
        "Customize representation to make it more concise."
        if self.action_set is not None:
            return f"P4IndirectAction(action_set={self.action_set!r})"
        if self.member_id is not None:
            return f"P4IndirectAction(member_id={self.member_id!r})"
        return f"P4IndirectAction(group_id={self.group_id!r})"


@dataclass(kw_only=True, slots=True)
class P4MeterConfig:
    """Represents a P4Runtime MeterConfig.

    Attributes:
        cir (int): Committed information rate (units/sec).
        cburst (int): Committed burst size.
        pir (int): Peak information rate (units/sec).
        pburst (int): Peak burst size.

    Example:
    ```
    config = P4MeterConfig(cir=10, cburst=20, pir=10, pburst=20)
    ```

    See Also:
        - P4TableEntry
        - P4MeterEntry
        - P4DirectMeterEntry
    """

    cir: int
    "Committed information rate (units/sec)."
    cburst: int
    "Committed burst size."
    pir: int
    "Peak information rate (units/sec)."
    pburst: int
    "Peak burst size."

    def encode(self) -> p4r.MeterConfig:
        "Encode object as MeterConfig."
        return p4r.MeterConfig(
            cir=self.cir, cburst=self.cburst, pir=self.pir, pburst=self.pburst
        )

    @classmethod
    def decode(cls, msg: p4r.MeterConfig) -> Self:
        "Decode MeterConfig."
        return cls(cir=msg.cir, cburst=msg.cburst, pir=msg.pir, pburst=msg.pburst)


@dataclass(kw_only=True, slots=True)
class P4CounterData:
    """Represents a P4Runtime object that keeps statistics of bytes and packets.

    Attributes:
        byte_count (int): the number of octets
        packet_count (int): the number of packets

    See Also:
        - P4TableEntry
        - P4MeterCounterData
        - P4CounterEntry
        - P4DirectCounterEntry
    """

    byte_count: int = 0
    "The number of octets."
    packet_count: int = 0
    "The number of packets."

    def encode(self) -> p4r.CounterData:
        "Encode object as CounterData."
        return p4r.CounterData(
            byte_count=self.byte_count, packet_count=self.packet_count
        )

    @classmethod
    def decode(cls, msg: p4r.CounterData) -> Self:
        "Decode CounterData."
        return cls(byte_count=msg.byte_count, packet_count=msg.packet_count)


@dataclass(kw_only=True, slots=True)
class P4MeterCounterData:
    """Represents a P4Runtime MeterCounterData that stores per-color counters.

    Attributes:
        green (CounterData): counter data for packets marked GREEN.
        yellow (CounterData): counter data for packets marked YELLOW.
        red (CounterData): counter data for packets marked RED.

    See Also:
        - P4TableEntry
        - P4MeterEntry
        - P4DirectMeterEntry
    """

    green: P4CounterData
    "Counter of packets marked GREEN."
    yellow: P4CounterData
    "Counter of packets marked YELLOW."
    red: P4CounterData
    "Counter of packets marked RED."

    def encode(self) -> p4r.MeterCounterData:
        "Encode object as MeterCounterData."
        return p4r.MeterCounterData(
            green=self.green.encode(),
            yellow=self.yellow.encode(),
            red=self.red.encode(),
        )

    @classmethod
    def decode(cls, msg: p4r.MeterCounterData) -> Self:
        "Decode MeterCounterData."
        return cls(
            green=P4CounterData.decode(msg.green),
            yellow=P4CounterData.decode(msg.yellow),
            red=P4CounterData.decode(msg.red),
        )


@decodable("table_entry")
@dataclass(slots=True)
class P4TableEntry(_P4Writable):
    """Represents a P4Runtime table entry.

    Attributes:
        table_id (str): Name of the table.
        match (P4TableMatch | None): Entry's match fields.
        action (P4TableAction | P4IndirectAction | None): Entry's action.
        is_default_action (bool): True if entry is the default table entry.
        priority (int): Priority of a table entry when match implies TCAM lookup.
        metadata (bytes): Arbitrary controller cookie (1.2.0).
        controller_metadata (int): Deprecated controller cookie (< 1.2.0).
        meter_config (P4MeterConfig | None): Meter configuration.
        counter_data (P4CounterData | None): Counter data for table entry.
        meter_counter_data (P4MeterCounterData | None): Meter counter data (1.4.0).
        idle_timeout_ns (int): Idle timeout in nanoseconds.
        time_since_last_hit (int | None): Nanoseconds since entry last matched.
        is_const (bool): True if entry is constant (1.4.0).

    The most commonly used fields are table_id, match, action, is_default_action,
    and priority. See the P4Runtime Spec for usage examples regarding the other
    attributes.

    When writing a P4TableEntry, you can specify the type of update using '+',
    '-', and '~'.

    Examples:
    ```
    # Specify all tables when using "read".
    entry = fy.P4TableEntry()

    # Specify the table named "ipv4" when using "read".
    entry = fy.P4TableEntry("ipv4")

    # Specify the default entry in the "ipv4" table when using "read".
    entry = fy.P4TableEntry("ipv4", is_default_action=True)

    # Insert an entry into the "ipv4" table.
    update = +fy.P4TableEntry(
        "ipv4",
        match=fy.Match(ipv4_dst="10.0.0.0/8"),
        action=fy.Action("forward", port=1),
    )

    # Modify the default action in the "ipv4" table.
    update = ~fy.P4TableEntry(
        "ipv4",
        action=fy.Action("forward", port=5),
        is_default_action=True
    )
    ```

    Operators:
        You can retrieve a match field from a table entry using `[]`. For
        example, `entry["ipv4_dst"]` is the same as `entry.match["ipv4_dst"]`.

    Formatting Helpers:
        The `match_str` and `action_str` methods provide P4Info-aware formatting
        of the match and action attributes.
    """

    table_id: str = ""
    "Name of the table."
    _: KW_ONLY
    match: P4TableMatch | None = None
    "Entry's match fields."
    action: P4TableAction | P4IndirectAction | None = None
    "Entry's action."
    is_default_action: bool = False
    "True if entry is the default table entry."
    priority: int = 0
    "Priority of a table entry when match implies TCAM lookup."
    metadata: bytes = b""
    "Arbitrary controller cookie. (1.2.0)."
    controller_metadata: int = 0
    "Deprecated controller cookie (< 1.2.0)."
    meter_config: P4MeterConfig | None = None
    "Meter configuration."
    counter_data: P4CounterData | None = None
    "Counter data for table entry."
    meter_counter_data: P4MeterCounterData | None = None
    "Meter counter data (1.4.0)."
    idle_timeout_ns: int = 0
    "Idle timeout in nanoseconds."
    time_since_last_hit: int | None = None
    "Nanoseconds since entry last matched."
    is_const: bool = False
    "True if entry is constant (1.4.0)."

    def __getitem__(self, key: str) -> Any:
        "Convenience accessor to retrieve a value from the `match` property."
        if self.match is not None:
            return self.match[key]
        raise KeyError(key)

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode TableEntry data as protobuf."
        return p4r.Entity(table_entry=self.encode_entry(schema))

    def encode_entry(self, schema: P4Schema) -> p4r.TableEntry:
        "Encode TableEntry data as protobuf."
        if not self.table_id:
            return self._encode_empty()

        table = schema.tables[self.table_id]

        if self.match:
            match = self.match.encode(table)
        else:
            match = None

        if self.action:
            action = self.action.encode_table_action(table)
        else:
            action = None

        if self.meter_config:
            meter_config = self.meter_config.encode()
        else:
            meter_config = None

        if self.counter_data:
            counter_data = self.counter_data.encode()
        else:
            counter_data = None

        if self.meter_counter_data:
            meter_counter_data = self.meter_counter_data.encode()
        else:
            meter_counter_data = None

        if self.time_since_last_hit is not None:
            time_since_last_hit = p4r.TableEntry.IdleTimeout(
                elapsed_ns=self.time_since_last_hit
            )
        else:
            time_since_last_hit = None

        return p4r.TableEntry(
            table_id=table.id,
            match=match,
            action=action,
            priority=self.priority,
            controller_metadata=self.controller_metadata,
            meter_config=meter_config,
            counter_data=counter_data,
            meter_counter_data=meter_counter_data,
            is_default_action=self.is_default_action,
            idle_timeout_ns=self.idle_timeout_ns,
            time_since_last_hit=time_since_last_hit,
            metadata=self.metadata,
            is_const=self.is_const,
        )

    def _encode_empty(self) -> p4r.TableEntry:
        "Encode an empty wildcard request."
        if self.counter_data is not None:
            counter_data = self.counter_data.encode()
        else:
            counter_data = None

        # FIXME: time_since_last_hit not supported for wildcard reads?
        if self.time_since_last_hit is not None:
            time_since_last_hit = p4r.TableEntry.IdleTimeout(
                elapsed_ns=self.time_since_last_hit
            )
        else:
            time_since_last_hit = None

        return p4r.TableEntry(
            counter_data=counter_data,
            time_since_last_hit=time_since_last_hit,
        )

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to TableEntry data."
        return cls.decode_entry(msg.table_entry, schema)

    @classmethod
    def decode_entry(cls, entry: p4r.TableEntry, schema: P4Schema) -> Self:
        "Decode protobuf to TableEntry data."
        if entry.table_id == 0:
            return cls("")

        table = schema.tables[entry.table_id]

        if entry.match:
            match = P4TableMatch.decode(entry.match, table)
        else:
            match = None

        if entry.HasField("action"):
            action = P4TableAction.decode_table_action(entry.action, table)
        else:
            action = None

        if entry.HasField("time_since_last_hit"):
            last_hit = entry.time_since_last_hit.elapsed_ns
        else:
            last_hit = None

        if entry.HasField("meter_config"):
            meter_config = P4MeterConfig.decode(entry.meter_config)
        else:
            meter_config = None

        if entry.HasField("counter_data"):
            counter_data = P4CounterData.decode(entry.counter_data)
        else:
            counter_data = None

        if entry.HasField("meter_counter_data"):
            meter_counter_data = P4MeterCounterData.decode(entry.meter_counter_data)
        else:
            meter_counter_data = None

        return cls(
            table_id=table.alias,
            match=match,
            action=action,
            priority=entry.priority,
            controller_metadata=entry.controller_metadata,
            meter_config=meter_config,
            counter_data=counter_data,
            meter_counter_data=meter_counter_data,
            is_default_action=entry.is_default_action,
            idle_timeout_ns=entry.idle_timeout_ns,
            time_since_last_hit=last_hit,
            metadata=entry.metadata,
            is_const=entry.is_const,
        )

    def match_dict(
        self,
        schema: P4Schema,
        *,
        wildcard: str | None = None,
    ) -> dict[str, str]:
        """Format the match fields as a dictionary of strings.

        If `wildcard` is None, only include match fields that have values. If
        `wildcard` is set, include all field names but replace unset values with
        given wildcard value (e.g. "*")
        """
        table = schema.tables[self.table_id]
        if self.match is not None:
            return self.match.format_dict(table, wildcard=wildcard)
        return P4TableMatch().format_dict(table, wildcard=wildcard)

    def match_str(
        self,
        schema: P4Schema,
        *,
        wildcard: str | None = None,
    ) -> str:
        "Format the match fields as a human-readable, canonical string."
        table = schema.tables[self.table_id]
        if self.match is not None:
            return self.match.format_str(table, wildcard=wildcard)
        return P4TableMatch().format_str(table, wildcard=wildcard)

    def action_str(self, schema: P4Schema) -> str:
        "Format the actions as a human-readable, canonical string."
        table = schema.tables[self.table_id]
        if self.action is None:
            return NOACTION_STR
        return self.action.format_str(table)


@decodable("register_entry")
@dataclass(slots=True)
class P4RegisterEntry(_P4ModifyOnly):
    "Represents a P4Runtime RegisterEntry."

    register_id: str = ""
    _: KW_ONLY
    index: int | None = None
    data: _DataValueType | None = None

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode RegisterEntry data as protobuf."
        if not self.register_id:
            return p4r.Entity(register_entry=p4r.RegisterEntry())

        register = schema.registers[self.register_id]

        if self.index is not None:
            index = p4r.Index(index=self.index)
        else:
            index = None

        if self.data is not None:
            data = register.type_spec.encode_data(self.data)
        else:
            data = None

        entry = p4r.RegisterEntry(
            register_id=register.id,
            index=index,
            data=data,
        )
        return p4r.Entity(register_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to RegisterEntry data."
        entry = msg.register_entry
        if entry.register_id == 0:
            return cls()

        register = schema.registers[entry.register_id]

        if entry.HasField("index"):
            index = entry.index.index
        else:
            index = None

        if entry.HasField("data"):
            data = register.type_spec.decode_data(entry.data)
        else:
            data = None

        return cls(
            register.alias,
            index=index,
            data=data,
        )


@decodable("multicast_group_entry")
@dataclass(slots=True)
class P4MulticastGroupEntry(_P4Writable):
    "Represents a P4Runtime MulticastGroupEntry."

    multicast_group_id: int = 0
    _: KW_ONLY
    replicas: Sequence[_ReplicaType] = ()
    metadata: bytes = b""

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode MulticastGroupEntry data as protobuf."
        entry = p4r.MulticastGroupEntry(
            multicast_group_id=self.multicast_group_id,
            replicas=[encode_replica(replica) for replica in self.replicas],
            metadata=self.metadata,
        )
        return p4r.Entity(
            packet_replication_engine_entry=p4r.PacketReplicationEngineEntry(
                multicast_group_entry=entry
            )
        )

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to MulticastGroupEntry data."
        entry = msg.packet_replication_engine_entry.multicast_group_entry
        return cls(
            multicast_group_id=entry.multicast_group_id,
            replicas=tuple(decode_replica(replica) for replica in entry.replicas),
            metadata=entry.metadata,
        )

    def replicas_str(self) -> str:
        "Format the replicas as a human-readable, canonical string."
        return " ".join(format_replica(rep) for rep in self.replicas)


@decodable("clone_session_entry")
@dataclass(slots=True)
class P4CloneSessionEntry(_P4Writable):
    "Represents a P4Runtime CloneSessionEntry."

    session_id: int = 0
    _: KW_ONLY
    class_of_service: int = 0
    packet_length_bytes: int = 0
    replicas: Sequence[_ReplicaType] = ()

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode CloneSessionEntry data as protobuf."
        entry = p4r.CloneSessionEntry(
            session_id=self.session_id,
            class_of_service=self.class_of_service,
            packet_length_bytes=self.packet_length_bytes,
            replicas=[encode_replica(replica) for replica in self.replicas],
        )
        return p4r.Entity(
            packet_replication_engine_entry=p4r.PacketReplicationEngineEntry(
                clone_session_entry=entry
            )
        )

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to CloneSessionEntry data."
        entry = msg.packet_replication_engine_entry.clone_session_entry
        return cls(
            session_id=entry.session_id,
            class_of_service=entry.class_of_service,
            packet_length_bytes=entry.packet_length_bytes,
            replicas=tuple(decode_replica(replica) for replica in entry.replicas),
        )

    def replicas_str(self) -> str:
        "Format the replicas as a human-readable, canonical string."
        return " ".join(format_replica(rep) for rep in self.replicas)


@decodable("digest_entry")
@dataclass(slots=True)
class P4DigestEntry(_P4Writable):
    "Represents a P4Runtime DigestEntry."

    digest_id: str = ""
    _: KW_ONLY
    max_list_size: int = 0
    max_timeout_ns: int = 0
    ack_timeout_ns: int = 0

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode DigestEntry data as protobuf."
        if not self.digest_id:
            return p4r.Entity(digest_entry=p4r.DigestEntry())

        digest = schema.digests[self.digest_id]

        if self.max_list_size == self.max_timeout_ns == self.ack_timeout_ns == 0:
            config = None
        else:
            config = p4r.DigestEntry.Config(
                max_timeout_ns=self.max_timeout_ns,
                max_list_size=self.max_list_size,
                ack_timeout_ns=self.ack_timeout_ns,
            )

        entry = p4r.DigestEntry(digest_id=digest.id, config=config)
        return p4r.Entity(digest_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to DigestEntry data."
        entry = msg.digest_entry
        if entry.digest_id == 0:
            return cls()

        digest = schema.digests[entry.digest_id]

        config = entry.config
        return cls(
            digest.alias,
            max_list_size=config.max_list_size,
            max_timeout_ns=config.max_timeout_ns,
            ack_timeout_ns=config.ack_timeout_ns,
        )


@decodable("action_profile_member")
@dataclass(slots=True)
class P4ActionProfileMember(_P4Writable):
    "Represents a P4Runtime ActionProfileMember."

    action_profile_id: str = ""
    _: KW_ONLY
    member_id: int = 0
    action: P4TableAction | None = None

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4ActionProfileMember as protobuf."
        if not self.action_profile_id:
            return p4r.Entity(action_profile_member=p4r.ActionProfileMember())

        profile = schema.action_profiles[self.action_profile_id]

        if self.action:
            action = self.action.encode_action(schema)
        else:
            action = None

        entry = p4r.ActionProfileMember(
            action_profile_id=profile.id,
            member_id=self.member_id,
            action=action,
        )
        return p4r.Entity(action_profile_member=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to ActionProfileMember data."
        entry = msg.action_profile_member
        if entry.action_profile_id == 0:
            return cls()

        profile = schema.action_profiles[entry.action_profile_id]

        if entry.HasField("action"):
            action = P4TableAction.decode_action(entry.action, schema)
        else:
            action = None

        return cls(
            action_profile_id=profile.alias,
            member_id=entry.member_id,
            action=action,
        )

    def action_str(self, schema: P4Schema) -> str:
        "Format the action as a human-readable, canonical string."
        if self.action is None:
            return NOACTION_STR
        return self.action.format_str(schema)


@dataclass(slots=True)
class P4Member:
    """Represents an ActionProfileGroup Member.

    See Also:
        - P4ActionProfileGroup
    """

    member_id: int
    _: KW_ONLY
    weight: P4Weight

    def encode(self) -> p4r.ActionProfileGroup.Member:
        "Encode P4Member as protobuf."
        match self.weight:
            case int(weight):
                watch_port = None
            case (int(weight), int(watch)):
                watch_port = encode_watch_port(watch)
            case other:  # pyright: ignore[reportUnnecessaryComparison]
                raise ValueError(f"unexpected weight: {other!r}")

        member = p4r.ActionProfileGroup.Member(
            member_id=self.member_id,
            weight=weight,
        )

        if watch_port is not None:
            member.watch_port = watch_port
        return member

    @classmethod
    def decode(cls, msg: p4r.ActionProfileGroup.Member) -> Self:
        "Decode protobuf to P4Member."
        match msg.WhichOneof("watch_kind"):
            case "watch_port":
                weight = (msg.weight, decode_watch_port(msg.watch_port))
            case None:
                weight = msg.weight
            case other:
                # "watch" (deprecated) is not supported
                raise ValueError(f"unknown oneof: {other!r}")

        return cls(member_id=msg.member_id, weight=weight)


@decodable("action_profile_group")
@dataclass(slots=True)
class P4ActionProfileGroup(_P4Writable):
    "Represents a P4Runtime ActionProfileGroup."

    action_profile_id: str = ""
    _: KW_ONLY
    group_id: int = 0
    max_size: int = 0
    members: Sequence[P4Member] | None = None

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4ActionProfileGroup as protobuf."
        if not self.action_profile_id:
            return p4r.Entity(action_profile_group=p4r.ActionProfileGroup())

        profile = schema.action_profiles[self.action_profile_id]

        if self.members is not None:
            members = [member.encode() for member in self.members]
        else:
            members = None

        entry = p4r.ActionProfileGroup(
            action_profile_id=profile.id,
            group_id=self.group_id,
            members=members,
            max_size=self.max_size,
        )
        return p4r.Entity(action_profile_group=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to ActionProfileGroup data."
        entry = msg.action_profile_group
        if entry.action_profile_id == 0:
            return cls()

        profile = schema.action_profiles[entry.action_profile_id]

        if entry.members:
            members = [P4Member.decode(member) for member in entry.members]
        else:
            members = None

        return cls(
            action_profile_id=profile.alias,
            group_id=entry.group_id,
            max_size=entry.max_size,
            members=members,
        )

    def action_str(self, _schema: P4Schema) -> str:
        "Return string representation of the weighted members."
        if not self.members:
            return ""

        return " ".join(
            [f"{member.weight}*{member.member_id:#x}" for member in self.members]
        )


@decodable("meter_entry")
@dataclass(slots=True)
class P4MeterEntry(_P4ModifyOnly):
    "Represents a P4Runtime MeterEntry."

    meter_id: str = ""
    _: KW_ONLY
    index: int | None = None
    config: P4MeterConfig | None = None
    counter_data: P4MeterCounterData | None = None

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4MeterEntry to protobuf."
        if not self.meter_id:
            return p4r.Entity(meter_entry=p4r.MeterEntry())

        meter = schema.meters[self.meter_id]

        if self.index is not None:
            index = p4r.Index(index=self.index)
        else:
            index = None

        if self.config is not None:
            config = self.config.encode()
        else:
            config = None

        if self.counter_data is not None:
            counter_data = self.counter_data.encode()
        else:
            counter_data = None

        entry = p4r.MeterEntry(
            meter_id=meter.id,
            index=index,
            config=config,
            counter_data=counter_data,
        )
        return p4r.Entity(meter_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to P4MeterEntry."
        entry = msg.meter_entry
        if not entry.meter_id:
            return cls()

        meter = schema.meters[entry.meter_id]

        if entry.HasField("index"):
            index = entry.index.index
        else:
            index = None

        if entry.HasField("config"):
            config = P4MeterConfig.decode(entry.config)
        else:
            config = None

        if entry.HasField("counter_data"):
            counter_data = P4MeterCounterData.decode(entry.counter_data)
        else:
            counter_data = None

        return cls(
            meter_id=meter.alias,
            index=index,
            config=config,
            counter_data=counter_data,
        )


@decodable("direct_meter_entry")
@dataclass(kw_only=True, slots=True)
class P4DirectMeterEntry(_P4ModifyOnly):
    "Represents a P4Runtime DirectMeterEntry."

    table_entry: P4TableEntry | None = None
    config: P4MeterConfig | None = None
    counter_data: P4MeterCounterData | None = None

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4DirectMeterEntry as protobuf."
        if self.table_entry is not None:
            table_entry = self.table_entry.encode_entry(schema)
        else:
            table_entry = None

        if self.config is not None:
            config = self.config.encode()
        else:
            config = None

        if self.counter_data is not None:
            counter_data = self.counter_data.encode()
        else:
            counter_data = None

        entry = p4r.DirectMeterEntry(
            table_entry=table_entry,
            config=config,
            counter_data=counter_data,
        )
        return p4r.Entity(direct_meter_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to P4DirectMeterEntry."
        entry = msg.direct_meter_entry

        if entry.HasField("table_entry"):
            table_entry = P4TableEntry.decode_entry(entry.table_entry, schema)
        else:
            table_entry = None

        if entry.HasField("config"):
            config = P4MeterConfig.decode(entry.config)
        else:
            config = None

        if entry.HasField("counter_data"):
            counter_data = P4MeterCounterData.decode(entry.counter_data)
        else:
            counter_data = None

        return cls(
            table_entry=table_entry,
            config=config,
            counter_data=counter_data,
        )


@decodable("counter_entry")
@dataclass(slots=True)
class P4CounterEntry(_P4ModifyOnly):
    "Represents a P4Runtime CounterEntry."

    counter_id: str = ""
    _: KW_ONLY
    index: int | None = None
    data: P4CounterData | None = None

    @property
    def packet_count(self) -> int:
        "Packet count from counter data (or 0 if there is no data)."
        if self.data is not None:
            return self.data.packet_count
        return 0

    @property
    def byte_count(self) -> int:
        "Byte count from counter data (or 0 if there is no data)."
        if self.data is not None:
            return self.data.byte_count
        return 0

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4CounterEntry as protobuf."
        if not self.counter_id:
            return p4r.Entity(counter_entry=p4r.CounterEntry())

        counter = schema.counters[self.counter_id]

        if self.index is not None:
            index = p4r.Index(index=self.index)
        else:
            index = None

        if self.data is not None:
            data = self.data.encode()
        else:
            data = None

        entry = p4r.CounterEntry(
            counter_id=counter.id,
            index=index,
            data=data,
        )
        return p4r.Entity(counter_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to P4CounterEntry."
        entry = msg.counter_entry
        if not entry.counter_id:
            return cls()

        counter = schema.counters[entry.counter_id]

        if entry.HasField("index"):
            index = entry.index.index
        else:
            index = None

        if entry.HasField("data"):
            data = P4CounterData.decode(entry.data)
        else:
            data = None

        return cls(counter_id=counter.alias, index=index, data=data)


@decodable("direct_counter_entry")
@dataclass(slots=True)
class P4DirectCounterEntry(_P4ModifyOnly):
    "Represents a P4Runtime DirectCounterEntry."

    counter_id: str = ""
    _: KW_ONLY
    table_entry: P4TableEntry | None = None
    data: P4CounterData | None = None

    @property
    def table_id(self) -> str:
        "Return table_id of related table."
        if self.table_entry is None:
            return ""
        return self.table_entry.table_id

    @property
    def packet_count(self) -> int:
        "Packet count from counter data (or 0 if there is no data)."
        if self.data is not None:
            return self.data.packet_count
        return 0

    @property
    def byte_count(self) -> int:
        "Byte count from counter data (or 0 if there is no data)."
        if self.data is not None:
            return self.data.byte_count
        return 0

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4DirectCounterEntry as protobuf."
        if self.table_entry is None:
            # Use `counter_id` to construct a `P4TableEntry` with the proper
            # table name.
            if self.counter_id:
                tb_name = schema.direct_counters[self.counter_id].direct_table_name
                table_entry = P4TableEntry(tb_name)
            else:
                table_entry = P4TableEntry()
        else:
            table_entry = self.table_entry

        if self.data is not None:
            data = self.data.encode()
        else:
            data = None

        entry = p4r.DirectCounterEntry(
            table_entry=table_entry.encode_entry(schema),
            data=data,
        )
        return p4r.Entity(direct_counter_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to P4DirectCounterEntry."
        entry = msg.direct_counter_entry

        if entry.HasField("table_entry"):
            table_entry = P4TableEntry.decode_entry(entry.table_entry, schema)
        else:
            table_entry = None

        if entry.HasField("data"):
            data = P4CounterData.decode(entry.data)
        else:
            data = None

        # Determine `counter_id` from table_entry.
        counter_id = ""
        if table_entry is not None and table_entry.table_id:
            direct_counter = schema.tables[table_entry.table_id].direct_counter
            assert direct_counter is not None
            counter_id = direct_counter.alias

        return cls(counter_id, table_entry=table_entry, data=data)


class P4ValueSetMember(dict[str, Any]):
    """Represents a sequence of P4Runtime FieldMatch in a ValueSet.

    To access an unnamed, singular value, use `member.value`.

    See Also:
        - P4ValueSetEntry
    """

    def __init__(
        self,
        __value: int | dict[str, Any] | None = None,
        **kwds: Any,
    ):
        if __value is None:
            super().__init__(**kwds)
        elif isinstance(__value, int):
            if kwds:
                raise ValueError("invalid keyword arguments")
            super().__init__({"": __value})
        else:
            super().__init__(__value, **kwds)

    @property
    def value(self) -> Any:
        "Return the unnamed, singular value."
        return self[""]

    def encode(self, value_set: P4ValueSet) -> list[p4r.FieldMatch]:
        "Encode P4ValueSetMember data as protobuf."
        result: list[p4r.FieldMatch] = []
        match = value_set.match

        for key, value in self.items():
            try:
                field = match[key].encode_field(value)
                if field is not None:
                    result.append(field)
            except Exception as ex:
                raise ValueError(
                    f"{value_set.name!r}: Match field {key!r}: {ex}"
                ) from ex

        return result

    @classmethod
    def decode(cls, msgs: Iterable[p4r.FieldMatch], value_set: P4ValueSet) -> Self:
        "Decode protobuf to P4ValueSetMember data."
        result: dict[str, Any] = {}
        match = value_set.match

        for field in msgs:
            fld = match[field.field_id]
            result[fld.alias] = fld.decode_field(field)

        return cls(result)


@decodable("value_set_entry")
@dataclass(slots=True)
class P4ValueSetEntry(_P4ModifyOnly):
    "Represents a P4Runtime ValueSetEntry."

    value_set_id: str
    _: KW_ONLY
    members: list[P4ValueSetMember]

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4ValueSetEntry as protobuf."
        value_set = schema.value_sets[self.value_set_id]
        members = [
            p4r.ValueSetMember(match=member.encode(value_set))
            for member in self.members
        ]

        return p4r.Entity(
            value_set_entry=p4r.ValueSetEntry(
                value_set_id=value_set.id, members=members
            )
        )

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to P4ValueSetEntry."
        entry = msg.value_set_entry
        value_set = schema.value_sets[entry.value_set_id]

        members = [
            P4ValueSetMember.decode(member.match, value_set) for member in entry.members
        ]

        return cls(value_set.alias, members=members)


@decodable("packet")
@dataclass(slots=True)
class P4PacketIn:
    "Represents a P4Runtime PacketIn."

    payload: bytes
    _: KW_ONLY
    metadata: _MetadataDictType

    @classmethod
    def decode(cls, msg: p4r.StreamMessageResponse, schema: P4Schema) -> Self:
        "Decode protobuf to PacketIn data."
        packet = msg.packet
        cpm = schema.controller_packet_metadata.get("packet_in")
        if cpm is None:
            # There is no controller metadata. Warn if message has any.
            pkt_meta = packet.metadata
            if pkt_meta:
                LOGGER.warning("P4PacketIn unexpected metadata: %r", pkt_meta)
            return cls(packet.payload, metadata={})

        return cls(
            packet.payload,
            metadata=cpm.decode(packet.metadata),
        )

    def __getitem__(self, key: str) -> Any:
        "Retrieve metadata value."
        return self.metadata[key]

    def __repr__(self) -> str:
        "Return friendlier hexadecimal description of packet."
        if self.metadata:
            return f"P4PacketIn(metadata={self.metadata!r}, payload=h'{self.payload.hex()}')"
        return f"P4PacketIn(payload=h'{self.payload.hex()}')"


@dataclass(slots=True)
class P4PacketOut:
    "Represents a P4Runtime PacketOut."

    payload: bytes
    _: KW_ONLY
    metadata: _MetadataDictType

    def __init__(self, __payload: bytes, /, **metadata: Any):
        self.payload = __payload
        self.metadata = metadata

    def encode_update(self, schema: P4Schema) -> p4r.StreamMessageRequest:
        "Encode PacketOut data as protobuf."
        cpm = schema.controller_packet_metadata["packet_out"]
        return p4r.StreamMessageRequest(
            packet=p4r.PacketOut(
                payload=self.payload,
                metadata=cpm.encode(self.metadata),
            )
        )

    def __getitem__(self, key: str) -> Any:
        "Retrieve metadata value."
        return self.metadata[key]

    def __repr__(self) -> str:
        "Return friendlier hexadecimal description of packet."
        if self.metadata:
            return f"P4PacketOut(metadata={self.metadata!r}, payload=h'{self.payload.hex()}')"
        return f"P4PacketOut(payload=h'{self.payload.hex()}')"


@decodable("digest")
@dataclass(slots=True)
class P4DigestList:
    "Represents a P4Runtime DigestList."

    digest_id: str
    _: KW_ONLY
    list_id: int
    timestamp: int
    data: list[_DataValueType]

    @classmethod
    def decode(cls, msg: p4r.StreamMessageResponse, schema: P4Schema) -> Self:
        "Decode protobuf to DigestList data."
        digest_list = msg.digest
        digest = schema.digests[digest_list.digest_id]

        type_spec = digest.type_spec
        return cls(
            digest_id=digest.alias,
            list_id=digest_list.list_id,
            timestamp=digest_list.timestamp,
            data=[type_spec.decode_data(item) for item in digest_list.data],
        )

    def __len__(self) -> int:
        "Return number of values in digest list."
        return len(self.data)

    def __getitem__(self, key: int) -> _DataValueType:
        "Retrieve value at given index from digest list."
        return self.data[key]

    def __iter__(self) -> Iterator[_DataValueType]:
        "Iterate over values in digest list."
        return iter(self.data)

    def ack(self) -> "P4DigestListAck":
        "Return the corresponding DigestListAck message."
        return P4DigestListAck(self.digest_id, self.list_id)


@dataclass(slots=True)
class P4DigestListAck:
    "Represents a P4Runtime DigestListAck."

    digest_id: str
    list_id: int

    def encode_update(self, schema: P4Schema) -> p4r.StreamMessageRequest:
        "Encode DigestListAck data as protobuf."
        digest = schema.digests[self.digest_id]

        return p4r.StreamMessageRequest(
            digest_ack=p4r.DigestListAck(
                digest_id=digest.id,
                list_id=self.list_id,
            )
        )


@decodable("idle_timeout_notification")
@dataclass(slots=True)
class P4IdleTimeoutNotification:
    "Represents a P4Runtime IdleTimeoutNotification."

    table_entry: list[P4TableEntry]
    timestamp: int

    @classmethod
    def decode(cls, msg: p4r.StreamMessageResponse, schema: P4Schema) -> Self:
        "Decode protobuf to IdleTimeoutNotification data."
        notification = msg.idle_timeout_notification
        table_entry = [
            P4TableEntry.decode_entry(entry, schema)
            for entry in notification.table_entry
        ]

        return cls(table_entry=table_entry, timestamp=notification.timestamp)

    def __len__(self) -> int:
        "Return number of table entries."
        return len(self.table_entry)

    def __getitem__(self, key: int) -> P4TableEntry:
        "Retrieve table entry at given index."
        return self.table_entry[key]

    def __iter__(self) -> Iterator[P4TableEntry]:
        "Iterate over table entries."
        return iter(self.table_entry)


@decodable("extern_entry")
@dataclass(kw_only=True, slots=True)
class P4ExternEntry(_P4Writable):
    "Represents a P4Runtime ExternEntry."

    extern_type_id: str
    extern_id: str
    entry: pbutil.PBAny

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode ExternEntry data as protobuf."
        extern = schema.externs[self.extern_type_id, self.extern_id]
        entry = p4r.ExternEntry(
            extern_type_id=extern.extern_type_id,
            extern_id=extern.id,
            entry=self.entry,
        )
        return p4r.Entity(extern_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to ExternEntry data."
        entry = msg.extern_entry
        extern = schema.externs[entry.extern_type_id, entry.extern_id]
        return cls(
            extern_type_id=extern.extern_type_name,
            extern_id=extern.name,
            entry=entry.entry,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gnmi1/gnmi_ext.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14gnmi1/gnmi_ext.proto\x12\x08gnmi_ext\x1a\x1egoogle/protobuf/duration.proto\"\xf2\x01\n\tExtension\x12\x37\n\x0eregistered_ext\x18\x01 \x01(\x0b\x32\x1d.gnmi_ext.RegisteredExtensionH\x00\x12\x39\n\x12master_arbitration\x18\x02 \x01(\x0b\x32\x1b.gnmi_ext.MasterArbitrationH\x00\x12$\n\x07history\x18\x03 \x01(\x0b\x32\x11.gnmi_ext.HistoryH\x00\x12\"\n\x06\x63ommit\x18\x04 \x01(\x0b\x32\x10.gnmi_ext.CommitH\x00\x12 \n\x05\x64\x65pth\x18\x05 \x01(\x0b\x32\x0f.gnmi_ext.DepthH\x00\x42\x05\n\x03\x65xt\"E\n\x13RegisteredExtension\x12!\n\x02id\x18\x01 \x01(\x0e\x32\x15.gnmi_ext.ExtensionID\x12\x0b\n\x03msg\x18\x02 \x01(\x0c\"Y\n\x11MasterArbitration\x12\x1c\n\x04role\x18\x01 \x01(\x0b\x32\x0e.gnmi_ext.Role\x12&\n\x0b\x65lection_id\x18\x02 \x01(\x0b\x32\x11.gnmi_ext.Uint128\"$\n\x07Uint128\x12\x0c\n\x04high\x18\x01 \x01(\x04\x12\x0b\n\x03low\x18\x02 \x01(\x04\"\x12\n\x04Role\x12\n\n\x02id\x18\x01 \x01(\t\"S\n\x07History\x12\x17\n\rsnapshot_time\x18\x01 \x01(\x03H\x00\x12$\n\x05range\x18\x02 \x01(\x0b\x32\x13.gnmi_ext.TimeRangeH\x00\x42\t\n\x07request\"\'\n\tTimeRange\x12\r\n\x05start\x18\x01 \x01(\x03\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x03\"\xe5\x01\n\x06\x43ommit\x12\n\n\x02id\x18\x01 \x01(\t\x12)\n\x06\x63ommit\x18\x02 \x01(\x0b\x32\x17.gnmi_ext.CommitRequestH\x00\x12*\n\x07\x63onfirm\x18\x03 \x01(\x0b\x32\x17.gnmi_ext.CommitConfirmH\x00\x12(\n\x06\x63\x61ncel\x18\x04 \x01(\x0b\x32\x16.gnmi_ext.CommitCancelH\x00\x12\x44\n\x15set_rollback_duration\x18\x05 \x01(\x0b\x32#.gnmi_ext.CommitSetRollbackDurationH\x00\x42\x08\n\x06\x61\x63tion\"E\n\rCommitRequest\x12\x34\n\x11rollback_duration\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\"\x0f\n\rCommitConfirm\"\x0e\n\x0c\x43ommitCancel\"Q\n\x19\x43ommitSetRollbackDuration\x12\x34\n\x11rollback_duration\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\"\x16\n\x05\x44\x65pth\x12\r\n\x05level\x18\x01 \x01(\r*3\n\x0b\x45xtensionID\x12\r\n\tEID_UNSET\x10\x00\x12\x15\n\x10\x45ID_EXPERIMENTAL\x10\xe7\x07\x42+Z)github.com/openconfig/gnmi/proto/gnmi_extb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gnmi1.gnmi_ext_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z)github.com/openconfig/gnmi/proto/gnmi_ext'
  _globals['_EXTENSIONID']._serialized_start=1100
  _globals['_EXTENSIONID']._serialized_end=1151
  _globals['_EXTENSION']._serialized_start=67
  _globals['_EXTENSION']._serialized_end=309
  _globals['_REGISTEREDEXTENSION']._serialized_start=311
  _globals['_REGISTEREDEXTENSION']._serialized_end=380
  _globals['_MASTERARBITRATION']._serialized_start=382
  _globals['_MASTERARBITRATION']._serialized_end=471
  _globals['_UINT128']._serialized_start=473
  _globals['_UINT128']._serialized_end=509
  _globals['_ROLE']._serialized_start=511
  _globals['_ROLE']._serialized_end=529
  _globals['_HISTORY']._serialized_start=531
  _globals['_HISTORY']._serialized_end=614
  _globals['_TIMERANGE']._serialized_start=616
  _globals['_TIMERANGE']._serialized_end=655
  _globals['_COMMIT']._serialized_start=658
  _globals['_COMMIT']._serialized_end=887
  _globals['_COMMITREQUEST']._serialized_start=889
  _globals['_COMMITREQUEST']._serialized_end=958
  _globals['_COMMITCONFIRM']._serialized_start=960
  _globals['_COMMITCONFIRM']._serialized_end=975
  _globals['_COMMITCANCEL']._serialized_start=977
  _globals['_COMMITCANCEL']._serialized_end=991
  _globals['_COMMITSETROLLBACKDURATION']._serialized_start=993
  _globals['_COMMITSETROLLBACKDURATION']._serialized_end=1074
  _globals['_DEPTH']._serialized_start=1076
  _globals['_DEPTH']._serialized_end=1098
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.duration_pb2
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class RetryInfo(google.protobuf.message.Message):
    """Describes when the clients can retry a failed request. Clients could ignore
    the recommendation here or retry when this information is missing from error
    responses.

    It's always recommended that clients should use exponential backoff when
    retrying.

    Clients should wait until `retry_delay` amount of time has passed since
    receiving the error response before retrying.  If retrying requests also
    fail, clients should use an exponential backoff scheme to gradually increase
    the delay between retries based on `retry_delay`, until either a maximum
    number of retries have been reached or a maximum retry delay cap has been
    reached.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RETRY_DELAY_FIELD_NUMBER: builtins.int
    @property
    def retry_delay(self) -> google.protobuf.duration_pb2.Duration:
        """Clients should wait at least this long between retrying the same request."""

    def __init__(
        self,
        *,
        retry_delay: google.protobuf.duration_pb2.Duration | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["retry_delay", b"retry_delay"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["retry_delay", b"retry_delay"]) -> None: ...

global___RetryInfo = RetryInfo

@typing.final
class DebugInfo(google.protobuf.message.Message):
    """Describes additional debugging info."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STACK_ENTRIES_FIELD_NUMBER: builtins.int
    DETAIL_FIELD_NUMBER: builtins.int
    detail: builtins.str
    """Additional debugging information provided by the server."""
    @property
    def stack_entries(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The stack trace entries indicating where the error occurred."""

    def __init__(
        self,
        *,
        stack_entries: collections.abc.Iterable[builtins.str] | None = ...,
        detail: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["detail", b"detail", "stack_entries", b"stack_entries"]) -> None: ...

global___DebugInfo = DebugInfo

@typing.final
class QuotaFailure(google.protobuf.message.Message):
    """Describes how a quota check failed.

    For example if a daily limit was exceeded for the calling project,
    a service could respond with a QuotaFailure detail containing the project
    id and the description of the quota limit that was exceeded.  If the
    calling project hasn't enabled the service in the developer console, then
    a service could respond with the project id and set `service_disabled`
    to true.

    Also see RetryInfo and Help types for other details about handling a
    quota failure.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Violation(google.protobuf.message.Message):
        """A message type used to describe a single quota violation.  For example, a
        daily quota or a custom quota that was exceeded.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SUBJECT_FIELD_NUMBER: builtins.int
        DESCRIPTION_FIELD_NUMBER: builtins.int
        subject: builtins.str
        """The subject on which the quota check failed.
        For example, "clientip:<ip address of client>" or "project:<Google
        developer project id>".
        """
        description: builtins.str
        """A description of how the quota check failed. Clients can use this
        description to find more about the quota configuration in the service's
        public documentation, or find the relevant quota limit to adjust through
        developer console.

        For example: "Service disabled" or "Daily Limit for read operations
        exceeded".
        """
        def __init__(
            self,
            *,
            subject: builtins.str = ...,
            description: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["description", b"description", "subject", b"subject"]) -> None: ...

    VIOLATIONS_FIELD_NUMBER: builtins.int
    @property
    def violations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___QuotaFailure.Violation]:
        """Describes all quota violations."""

    def __init__(
        self,
        *,
        violations: collections.abc.Iterable[global___QuotaFailure.Violation] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["violations", b"violations"]) -> None: ...

global___QuotaFailure = QuotaFailure

@typing.final
class ErrorInfo(google.protobuf.message.Message):
    """Describes the cause of the error with structured details.

    Example of an error when contacting the "pubsub.googleapis.com" API when it
    is not enabled:
        { "reason":   "API_DISABLED"
          "domain": "googleapis.com"
          "metadata": {
            "resource": "projects/123",
            "service": "pubsub.googleapis.com"
          }
        }
    This response indicates that the pubsub.googleapis.com API is not enabled.

    Example of an error that is returned when attempting to create a Spanner
    instance in a region that is out of stock:
        { "reason":   "STOCKOUT"
          "domain": "spanner.googleapis.com",
          "metadata": {
            "availableRegions": "us-central1,us-east2"
          }
        }
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class MetadataEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    REASON_FIELD_NUMBER: builtins.int
    DOMAIN_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    reason: builtins.str
    """The reason of the error. This is a constant value that identifies the
    proximate cause of the error. Error reasons are unique within a particular
    domain of errors. This should be at most 63 characters and match
    /[A-Z0-9_]+/.
    """
    domain: builtins.str
    """The logical grouping to which the "reason" belongs.  Often "domain" will
    contain the registered service name of the tool or product that is the
    source of the error. Example: "pubsub.googleapis.com". If the error is
    common across many APIs, the first segment of the example above will be
    omitted.  The value will be, "googleapis.com".
    """
    @property
    def metadata(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """Additional structured details about this error.

        Keys should match /[a-zA-Z0-9-_]/ and be limited to 64 characters in
        length. When identifying the current value of an exceeded limit, the units
        should be contained in the key, not the value.  For example, rather than
        {"instanceLimit": "100/request"}, should be returned as,
        {"instanceLimitPerRequest": "100"}, if the client exceeds the number of
        instances that can be created in a single (batch) request.
        """

    def __init__(
        self,
        *,
        reason: builtins.str = ...,
        domain: builtins.str = ...,
        metadata: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["domain", b"domain", "metadata", b"metadata", "reason", b"reason"]) -> None: ...

global___ErrorInfo = ErrorInfo

@typing.final
class PreconditionFailure(google.protobuf.message.Message):
    """Describes what preconditions have failed.

    For example, if an RPC failed because it required the Terms of Service to be
    acknowledged, it could list the terms of service violation in the
    PreconditionFailure message.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Violation(google.protobuf.message.Message):
        """A message type used to describe a single precondition failure."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TYPE_FIELD_NUMBER: builtins.int
        SUBJECT_FIELD_NUMBER: builtins.int
        DESCRIPTION_FIELD_NUMBER: builtins.int
        type: builtins.str
        """The type of PreconditionFailure. We recommend using a service-specific
        enum type to define the supported precondition violation subjects. For
        example, "TOS" for "Terms of Service violation".
        """
        subject: builtins.str
        """The subject, relative to the type, that failed.
        For example, "google.com/cloud" relative to the "TOS" type would indicate
        which terms of service is being referenced.
        """
        description: builtins.str
        """A description of how the precondition failed. Developers can use this
        description to understand how to fix the failure.

        For example: "Terms of service not accepted".
        """
        def __init__(
            self,
            *,
            type: builtins.str = ...,
            subject: builtins.str = ...,
            description: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["description", b"description", "subject", b"subject", "type", b"type"]) -> None: ...

    VIOLATIONS_FIELD_NUMBER: builtins.int
    @property
    def violations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___PreconditionFailure.Violation]:
        """Describes all precondition violations."""

    def __init__(
        self,
        *,
        violations: collections.abc.Iterable[global___PreconditionFailure.Violation] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["violations", b"violations"]) -> None: ...

global___PreconditionFailure = PreconditionFailure

@typing.final
class BadRequest(google.protobuf.message.Message):
    """Describes violations in a client request. This error type focuses on the
    syntactic aspects of the request.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class FieldViolation(google.protobuf.message.Message):
        """A message type used to describe a single bad request field."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        FIELD_FIELD_NUMBER: builtins.int
        DESCRIPTION_FIELD_NUMBER: builtins.int
        field: builtins.str
        """A path leading to a field in the request body. The value will be a
        sequence of dot-separated identifiers that identify a protocol buffer
        field. E.g., "field_violations.field" would identify this field.
        """
        description: builtins.str
        """A description of why the request element is bad."""
        def __init__(
            self,
            *,
            field: builtins.str = ...,
            description: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["description", b"description", "field", b"field"]) -> None: ...

    FIELD_VIOLATIONS_FIELD_NUMBER: builtins.int
    @property
    def field_violations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___BadRequest.FieldViolation]:
        """Describes all violations in a client request."""

    def __init__(
        self,
        *,
        field_violations: collections.abc.Iterable[global___BadRequest.FieldViolation] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["field_violations", b"field_violations"]) -> None: ...

global___BadRequest = BadRequest

@typing.final
class RequestInfo(google.protobuf.message.Message):
    """Contains metadata about the request that clients can attach when filing a bug
    or providing other forms of feedback.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REQUEST_ID_FIELD_NUMBER: builtins.int
    SERVING_DATA_FIELD_NUMBER: builtins.int
    request_id: builtins.str
    """An opaque string that should only be interpreted by the service generating
    it. For example, it can be used to identify requests in the service's logs.
    """
    serving_data: builtins.str
    """Any data that was used to serve this request. For example, an encrypted
    stack trace that can be sent back to the service provider for debugging.
    """
    def __init__(
        self,
        *,
        request_id: builtins.str = ...,
        serving_data: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["request_id", b"request_id", "serving_data", b"serving_data"]) -> None: ...

global___RequestInfo = RequestInfo

@typing.final
class ResourceInfo(google.protobuf.message.Message):
    """Describes the resource that is being accessed."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RESOURCE_TYPE_FIELD_NUMBER: builtins.int
    RESOURCE_NAME_FIELD_NUMBER: builtins.int
    OWNER_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    resource_type: builtins.str
    """A name for the type of resource being accessed, e.g. "sql table",
    "cloud storage bucket", "file", "Google calendar"; or the type URL
    of the resource: e.g. "type.googleapis.com/google.pubsub.v1.Topic".
    """
    resource_name: builtins.str
    """The name of the resource being accessed.  For example, a shared calendar
    name: "example.com_4fghdhgsrgh@group.calendar.google.com", if the current
    error is [google.rpc.Code.PERMISSION_DENIED][google.rpc.Code.PERMISSION_DENIED].
    """
    owner: builtins.str
    """The owner of the resource (optional).
    For example, "user:<owner email>" or "project:<Google developer project
    id>".
    """
    description: builtins.str
    """Describes what error is encountered when accessing this resource.
    For example, updating a cloud project may require the `writer` permission
    on the developer console project.
    """
    def __init__(
        self,
        *,
        resource_type: builtins.str = ...,
        resource_name: builtins.str = ...,
        owner: builtins.str = ...,
        description: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["description", b"description", "owner", b"owner", "resource_name", b"resource_name", "resource_type", b"resource_type"]) -> None: ...

global___ResourceInfo = ResourceInfo

@typing.final
class Help(google.protobuf.message.Message):
    """Provides links to documentation or for performing an out of band action.

    For example, if a quota check failed with an error indicating the calling
    project hasn't enabled the accessed service, this can contain a URL pointing
    directly to the right place in the developer console to flip the bit.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Link(google.protobuf.message.Message):
        """Describes a URL link."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        DESCRIPTION_FIELD_NUMBER: builtins.int
        URL_FIELD_NUMBER: builtins.int
        description: builtins.str
        """Describes what the link offers."""
        url: builtins.str
        """The URL of the link."""
        def __init__(
            self,
            *,
            description: builtins.str = ...,
            url: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["description", b"description", "url", b"url"]) -> None: ...

    LINKS_FIELD_NUMBER: builtins.int
    @property
    def links(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Help.Link]:
        """URL(s) pointing to additional information on handling the current error."""

    def __init__(
        self,
        *,
        links: collections.abc.Iterable[global___Help.Link] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["links", b"links"]) -> None: ...

global___Help = Help

@typing.final
class LocalizedMessage(google.protobuf.message.Message):
    """Provides a localized error message that is safe to return to the user
    which can be attached to an RPC error.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LOCALE_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    locale: builtins.str
    """The locale used following the specification defined at
    http://www.rfc-editor.org/rfc/bcp/bcp47.txt.
    Examples are: "en-US", "fr-CH", "es-MX"
    """
    message: builtins.str
    """The localized error message in the above locale."""
    def __init__(
        self,
        *,
        locale: builtins.str = ...,
        message: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["locale", b"locale", "message", b"message"]) -> None: ...

global___LocalizedMessage = LocalizedMessage
